#!/usr/bin/env python

import json
from pathlib import Path

import click
import yaml
from minio import Minio
import datetime
import json
from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from textwrap import dedent
from omegaconf import OmegaConf

import MAIA
version = MAIA.__version__

TIMESTAMP = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.datetime.now())

DESC = dedent(
    """
    Script to deploy the JupyterHub helm chart to a Kubernetes cluster. The target cluster is specified by setting the corresponding ``--cluster--config-file``,
    while the namespace-related configuration is specified with ``--form``.
    """  # noqa: E501
)
EPILOG = dedent(
    """
    Example call:
    ::
        {filename}  --form /PATH/TO/form.yaml --cluster-config-file /PATH/TO/cluster.yaml
    """.format(  # noqa: E501
        filename=Path(__file__).stem
    )
)

def get_arg_parser():
    pars = ArgumentParser(description=DESC, epilog=EPILOG, formatter_class=RawTextHelpFormatter)

    pars.add_argument(
        "--form",
        type=str,
        required=True,
        help="YAML configuration file used to extract the namespace configuration.",
    )
    pars.add_argument(
        "--maia-config-file",
        type=str,
        required=True,
        help="YAML configuration file used to extract the MAIA configuration.",
    )

    pars.add_argument(
        "--cluster-config-file",
        type=str,
        required=True,
        help="YAML configuration file used to extract the cluster configuration.",
    )

    pars.add_argument('-v', '--version', action='version', version='%(prog)s ' + version)

    return pars


@click.command()
@click.option("--form", type=str)
@click.option("--maia-config-file", type=str)
@click.option("--cluster-config-file", type=str)
def create_jupyterhub_config(form,maia_config_file,cluster_config_file):
    create_jupyterhub_config_api(form,maia_config_file,cluster_config_file)

def create_jupyterhub_config_api( form,
                                 maia_config_file,
                              cluster_config_file,
                                
                                  config_folder = None
                                  
                             ):

    if type(maia_config_file) == dict:
        maia_form = maia_config_file
    else:
        with open(maia_config_file, "r") as f:
            maia_form = yaml.safe_load(f)

    if type(cluster_config_file) == dict:
        cluster_config = cluster_config_file
    else:
        with open(cluster_config_file, "r") as f:
            cluster_config = yaml.safe_load(f)
    
    if type(form) == dict:
        user_form = form
    else:
        with open(form, "r") as f:
            user_form = yaml.safe_load(f)

    storage_class = cluster_config["storage_class"]

    traefik_resolver = None
    if "traefik_resolver" in cluster_config:
        traefik_resolver = cluster_config["traefik_resolver"]

    nginx_cluster_issuer = None
    if "nginx_cluster_issuer" in cluster_config:
        nginx_cluster_issuer = cluster_config["nginx_cluster_issuer"]

    hub_storage_class = None
    if "hub_storage_class" in cluster_config:
        hub_storage_class = cluster_config["hub_storage_class"]

    hub_image = None
    hub_tag = None
    if "hub_image" in cluster_config:
        hub_image = cluster_config["hub_image"]
        hub_tag = cluster_config["hub_tag"]

    base_url = None
    if "base_url" in cluster_config:
        base_url = cluster_config["base_url"]

    keycloak = None
    if "keycloak" in cluster_config:
        keycloak = cluster_config["keycloak"]

    namespace = user_form["group_ID"].lower().replace("_", "-")
    group_subdomain = user_form["group_subdomain"]
    team_id = user_form["group_ID"]
    resources_limits = user_form["resources_limits"]

    gpu_request = False
    if "gpu_request" in user_form:
        gpu_request = user_form["gpu_request"]

    domain = cluster_config["domain"]

    if "url_type" in cluster_config:
        if cluster_config["url_type"] == "subdomain":
            hub_address = f"{group_subdomain}.{domain}"
        elif cluster_config["url_type"] == "subpath":
            hub_address = domain
        else:
            hub_address = None
            
    admins = cluster_config.get("admins", [])


    ## Used for CIFS mount
    extra_host_volumes = [

    ]


    extra_volumes = [
       {
            "name": "jupyterhub-shared",
            "pvc-name": "shared",
            "mount-path": "/home/maia-user/shared"
        }
    ]

    jh_helm_template = {
      "resource": {
        "helm_release": {
          "jupyterhub": {
            "name": "jupyterhub-{}".format(namespace.lower()),
            "repository": "https://hub.jupyter.org/helm-chart/",
            "chart": "jupyterhub",
            "version": "3.1.0",
            "namespace": namespace.lower(),
            "create_namespace": False

          }
        }
      }
    }

    jh_template = {
        "cull": {
            "enabled": False
        },
        "ingress":  {
            "enabled": True,
            "hosts": [
                hub_address
            ],
            "annotations": {

            },
            "tls": [
                {
                    "hosts": [
                        hub_address
                    ]

                }
            ]
        },
        "hub":{
            # activeServerLimit  ## TODO: Add User Limit
            # concurrentSpawnLimit
        #"loadRoles": {
       #
        #    "user": {
        #    "description": "Allow users to access the shared server in addition to default perms",
        #    "scopes": ["self", "access:servers!user=<USER_EMAIL>"],
        #        "users": [""],
        #        "groups": [""]
        #}  ## TODO: Add Load Roles
        #                        },
            "config":{
                "GenericOAuthenticator":{

                    "login_service": "MAIA Account",
                    "username_claim": "preferred_username",
                    "scope": [
                        "openid",
                        "profile",
                        "email"
                    ],
                    "userdata_params": {
                        "state": "state"
                    },
                    "claim_groups_key": "groups",
                    "allowed_groups": [
                        f"MAIA:{team_id}"
                    ],
                    "admin_groups": [
                        "MAIA:admin"
                    ],

                },
                "JupyterHub":{
                    "admin_access": True,
                    "authenticator_class": "generic-oauth"
                },
                "Authenticator":{
                    "admin_users": admins,
                    "allowed_users": user_form["users"]
            }
        },


    },
        "singleuser":{
                "allowPrivilegeEscalation": True,
                "uid": 1000,
                "networkPolicy":{
                "enabled": False},
                "defaultUrl": "/lab/tree/Welcome.ipynb",
                "extraEnv":{
                    "GRANT_SUDO": "yes",
                    "SHELL": "/usr/bin/bash",
                    "TZ": "UTC",
                    "SIZEW": "1920",
                    "SIZEH": "1080",
                    "REFRESH": "60",
                    "DPI": "96",
                    "CDEPTH": "24",
                    "PASSWD": "maia",
                    "WEBRTC_ENCODER": "nvh264enc",
                    "BASIC_AUTH_PASSWORD": "maia",
                    "NOVNC_ENABLE": "true",
                    "ssh_publickey": "NOKEY",
                    "NB_USER": "maia-user",
                    "MINIO_ACCESS_KEY": user_form.get("minio_access_key", "N/A"),
                    "MINIO_SECRET_KEY": user_form.get("minio_secret_key", "N/A"),
                    "MLFLOW_TRACKING_URI": f"https://{hub_address}/mlflow",
                    "HOSTNAME": cluster_config["ssh_hostname"],
                    "NAMESPACE": namespace.lower()

                }
            }
    }

    if cluster_config["url_type"] == "subpath":
        jh_template["singleuser"]["extraEnv"]["MLFLOW_TRACKING_URI"] = f"https://{hub_address}/{namespace}-mlflow"

    if "minio_env_name" in user_form:
        minio_env_name = user_form["minio_env_name"]
        client = Minio(cluster_config["minio_url"],
                       access_key=cluster_config["minio_access_key"],
                       secret_key=cluster_config["minio_secret_key"],
                       secure=True)
        client.fget_object(cluster_config["bucket_name"], minio_env_name, minio_env_name)
        with open(minio_env_name, "r") as f:
            file_string = f.read()
            if file_string.startswith("name:"):
                jh_template["singleuser"]["extraEnv"]["CONDA_ENV"] = str(file_string)
            else:
                jh_template["singleuser"]["extraEnv"]["PIP_ENV"] = str(file_string)


    if "url_type" in cluster_config:
        if cluster_config["url_type"] == "subpath":
            jh_template["hub"]["baseUrl"] = f"/{group_subdomain}-hub"


    if keycloak is not None:
        jh_template["hub"]["config"]["GenericOAuthenticator"]["client_id"] = keycloak["client_id"]
        jh_template["hub"]["config"]["GenericOAuthenticator"]["client_secret"] = keycloak["client_secret"]
        jh_template["hub"]["config"]["GenericOAuthenticator"]["authorize_url"] = keycloak["authorize_url"]
        jh_template["hub"]["config"]["GenericOAuthenticator"]["token_url"] = keycloak["token_url"]
        jh_template["hub"]["config"]["GenericOAuthenticator"]["userdata_url"] = keycloak["userdata_url"]
        if "url_type" in cluster_config:
            if cluster_config["url_type"]  == "subdomain" :
                jh_template["hub"]["config"]["GenericOAuthenticator"]["oauth_callback_url"] = f"https://{hub_address}/hub/oauth_callback"
                # print("Register Callback: ")
                # print(f"https://{hub_address}/hub/oauth_callback")
            elif cluster_config["url_type"]  == "subpath":
                jh_template["hub"]["config"]["GenericOAuthenticator"][
                    "oauth_callback_url"] = f"https://{hub_address}/{group_subdomain}-hub/oauth_callback"
                # print("Register Callback: ")
                #print(f"https://{hub_address}/{group_subdomain}-hub/oauth_callback")


    #if "private_hub" in user_form:

    #    jh_template["hub"]["image"] = {

    #                "name": "registry.maia.cloud.cbh.kth.se/jupyterhub",
    #                "tag": "1.1"

    #    }

    if not gpu_request:
        jh_template["singleuser"]["extraEnv"]["NVIDIA_VISIBLE_DEVICES"] = ""


    if "ssh_users" in user_form:
        for ssh_port in user_form["ssh_users"]:
            username = ssh_port["username"].replace("@","__at__")
            jh_template["singleuser"]["extraEnv"][f"SSH_PORT_{username}"] = str(ssh_port["ssh_port"])


    if traefik_resolver is not None:
        jh_template["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = traefik_resolver
        jh_template["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"]= "websecure"
        jh_template["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"]= "true"

    if nginx_cluster_issuer is not None:
        jh_template["ingress"]["annotations"]["nginx.ingress.kubernetes.io/proxy-body-size"]= "2g"
        jh_template["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = nginx_cluster_issuer
        jh_template["ingress"]["tls"][0]["secretName"] = "jupyterhub-{}-tls".format(namespace.lower())

    if hub_storage_class is not None:
        jh_template["hub"]["db"] = {"pvc":{"storageClassName": hub_storage_class}}


    if hub_image is not None:
        jh_template["hub"]["image"] = {
            "name": hub_image,
            "tag": hub_tag
        }

    if base_url is not None:
        jh_template["hub"]["base_url"] = base_url


    #jh_template["singleuser"]["nodeSelector"] = {"kubernetes.io/hostname": "node-1"}  #TODO: Add nodeSelector

    jh_template["singleuser"]["storage"] = {
        "homeMountPath": "/home/maia-user",
        "dynamic": {
            "storageClass": storage_class
        },
        "extraVolumes": [
            {
                "name": "shm-volume",
                "emptyDir": {
                    "medium": "Memory"
                }
            }
        ],
        "extraVolumeMounts": [
            {
                "name": "shm-volume",
                "mountPath": "/dev/shm"
            }
        ]
    }


    if resources_limits["memory"][1].endswith(" Gi"):
        resources_limits["memory"][1] = resources_limits["memory"][1].replace(" Gi", "G")
    if resources_limits["memory"][0].endswith(" Gi"):
        resources_limits["memory"][0] = resources_limits["memory"][0].replace(" Gi", "G")

    jh_template["singleuser"]["memory"] = {
        "limit": resources_limits["memory"][1],
        "guarantee": resources_limits["memory"][0]
    }


    jh_template["singleuser"]["cpu"] = {
        "limit": int(resources_limits["cpu"][1]),
        "guarantee": int(resources_limits["cpu"][0])
    }

    for extra_volume in extra_volumes:
        jh_template["singleuser"]["storage"]["extraVolumes"].append({
            "name": extra_volume["name"],
            "persistentVolumeClaim": {
                "claimName": extra_volume["pvc-name"]
            }
        })
        jh_template["singleuser"]["storage"]["extraVolumeMounts"].append({
            "name": extra_volume["name"],
            "mountPath": extra_volume["mount-path"]
        })
    
    for extra_host_volume in extra_host_volumes:
        jh_template["singleuser"]["storage"]["extraVolumes"].append({
            "name": extra_host_volume["name"],
            "hostPath": {
                "path": extra_host_volume["host-path"]
            }
        })
        jh_template["singleuser"]["storage"]["extraVolumeMounts"].append({
            "name": extra_host_volume["name"],
            "mountPath": extra_host_volume["mount-path"]
        })

    jh_template["singleuser"]["image"] = {
        "name": "jupyter/datascience-notebook",
        "tag": "latest",
        
    }
    
    if "imagePullSecrets" in cluster_config:
        jh_template["singleuser"]["image"]["pullSecrets"] = cluster_config["imagePullSecrets"]

    maia_workspace_version = maia_form["maia_workspace_version"]
    maia_workspace_image = maia_form["maia_workspace_image"]
    jh_template["singleuser"]["profileList"] = [
        {"display_name": f"MAIA Workspace v{maia_workspace_version}",
         "description": "MAIA Workspace with Python 3.10, Anaconda, MatLab, RStudio, VSCode and SSH Connection",
         "default": True,
         "kubespawner_override":{"image": f"{maia_workspace_image}:{maia_workspace_version}",
                                "start_timeout": 3600,
                                "http_timeout": 3600,
                                #mem_limit
                                #cpu_limit
                                #mem_guarantee
                                #cpu_guarantee
                                #nodeSelector: {"kubernetes.io/hostname": "node-1"}
                               "extra_resource_limits": {
                               },

                                #"container_security_context": {
                                    #"privileged": True,
                                    #"procMount": "unmasked",
                                    #"seccompProfile": {
                                    #    "type": "Unconfined"
                                    #}
                                #}
                                }
         }

    ]

    if "maia_monai_toolkit_image" in maia_form:
        jh_template["singleuser"]["profileList"].append(
            {"display_name": "MONAI Toolkit 3.0",
            "description": "MONAI Toolkit 3.0, including MONAI Bundles from the MONAI Model ZOO and Tutorial Notebooks for MONAI Core, MONAI Label and MONAI FL",
            "kubespawner_override": {
                "image": maia_form["maia_monai_toolkit_image"],
                "start_timeout": 3600,
                "http_timeout": 3600,
                "extra_resource_limits": {
                    #"nvidia.com/gpu": "1"
                },
                "uid": 0,


            }
            })

    if gpu_request:
        jh_template["singleuser"]["profileList"][0]["kubespawner_override"]["extra_resource_limits"] = {
            "nvidia.com/gpu": "1"
        }




    jh_helm_template["resource"]["helm_release"]["jupyterhub"]["values"] = [
        yaml.dump(jh_template)
    ]

    
    jh_template["chart_name"] = jh_helm_template["resource"]["helm_release"]["jupyterhub"]["chart"]
    jh_template["chart_version"] = jh_helm_template["resource"]["helm_release"]["jupyterhub"]["version"]
    jh_template["repo_url"] = jh_helm_template["resource"]["helm_release"]["jupyterhub"]["repository"]

    Path(config_folder).joinpath(team_id,"jupyterhub_values").mkdir(parents=True, exist_ok=True)

    with open( Path(config_folder).joinpath(team_id,"jupyterhub_values","jupyterhub_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(jh_template))
    
    
    return {  ##  TODO: Causing issues from ArgoCD
        "namespace": namespace,
        "chart": jh_template["chart_name"],
        "release": f"{namespace}-jupyterhub",
        "repo": jh_template["repo_url"],
        "version": jh_template["chart_version"],
        "values": str(Path(config_folder).joinpath(team_id,"jupyterhub_values","jupyterhub_values.yaml"))

    }


def main():
    create_jupyterhub_config()

if __name__ == "__main__":
    main()




from pathlib import Path
import click
import yaml
import datetime
import json

import os
from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from textwrap import dedent

import MAIA
import random
import string
version = MAIA.__version__


TIMESTAMP = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.datetime.now())

DESC = dedent(
    """
    Script to initialize the MAIA Cluster configuration. The cluster configuration is specified by the ``--cluster-config`` option,
    and the configuration folder where the resolved configuration files will be saved is specified by the ``--config-folder`` option.
    """  # noqa: E501
)
EPILOG = dedent(
    """
    Example call:
    ::
        {filename} --cluster-config /PATH/TO/cluster_config.yaml --config-folder /PATH/TO/config_folder
    """.format(  # noqa: E501
        filename=Path(__file__).stem
    )
)

def get_arg_parser():
    pars = ArgumentParser(description=DESC, epilog=EPILOG, formatter_class=RawTextHelpFormatter)

    pars.add_argument(
        "--cluster-config",
        type=str,
        required=True,
        help="YAML configuration file used to extract the cluster configuration.",
    )

    pars.add_argument(
        "--config-folder",
        type=str,
        required=True,
        help="Configuration Folder where to save the resolved MAIA Cluster configuration files.",
    )

    pars.add_argument('-v', '--version', action='version', version='%(prog)s ' + version)

    return pars


def generate_random_password(length=12):
    characters = string.ascii_letters + string.digits
    return ''.join(random.choice(characters) for i in range(length))


@click.command()
@click.option("--cluster-config", type=str)
@click.option("--config-folder", type=str)
def main(cluster_config, config_folder):
    create_configuration(cluster_config, config_folder)

def create_configuration(cluster_config, config_folder):

   
    cluster_config_dict = yaml.safe_load(Path(cluster_config).read_text())



    cluster_config_dict["traefik_dashboard_password"] = generate_random_password(12)
    cluster_config_dict["rancher_password"] = generate_random_password(12)
    cluster_config_dict["ingress_class"] = "maia-core-traefik" #Alternative: "nginx"
    cluster_config_dict["traefik_resolver"] = "maiaresolver"
    cluster_config_dict["ssh_port_type"] = "NodePort"
    cluster_config_dict["port_range"] = [30000, 32767]
    cluster_config_dict["shared_storage_class"] = "microk8s-hostpath"
    cluster_config_dict["storage_class"] = "microk8s-hostpath"
    cluster_config_dict["url_type"] = "subdomain"
    cluster_config_dict["ssh_hostname"] = cluster_config_dict["domain"]
    cluster_config_dict["keycloak_maia_client_secret"] = generate_random_password(16)
    cluster_config_dict["keycloak"] = {
        "client_id": "maia",
        "issuer_url": f"https://iam.{cluster_config_dict['domain']}/realms/maia",
        "client_secret": cluster_config_dict["keycloak_maia_client_secret"],
        "authorize_url": f"https://iam.{cluster_config_dict['domain']}/realms/maia/protocol/openid-connect/auth",
        "token_url": f"https://iam.{cluster_config_dict['domain']}/realms/maia/protocol/openid-connect/token",
        "userdata_url": f"https://iam.{cluster_config_dict['domain']}/realms/maia/protocol/openid-connect/userinfo"
    }
    cluster_config_dict["argocd_destination_cluster_address"] = f"https://mgmt.{cluster_config_dict['domain']}/k8s/clusters/local"
    cluster_config_dict["services"] = {
        "dashboard": f"https://dashboard.{cluster_config_dict['domain']}",
        "login":  f"https://login.{cluster_config_dict['domain']}",
        "grafana": f"https://grafana.{cluster_config_dict['domain']}",
        "rancher": f"https://mgmt.{cluster_config_dict['domain']}",
        "argocd": f"https://argocd.{cluster_config_dict['domain']}",
        "registry": f"https://registry.{cluster_config_dict['domain']}",
        "keycloak": f"https://iam.{cluster_config_dict['domain']}",
        "traefik": f"https://traefik.{cluster_config_dict['domain']}"

    }

    cluster_config_dict["maia_dashboard"] = {
        "enabled": True,
        "token": ""
    }


    with open(os.path.join(config_folder,"MAIA_realm_template.json"), "r") as f:
        realm = json.load(f)
        clients = realm["clients"]
        for idx,client in enumerate(clients):
            if client['clientId'] == 'maia':
                clients[idx]['secret'] = cluster_config_dict["keycloak_maia_client_secret"]
    
    with open(os.path.join(config_folder,"MAIA_realm.json"), "w") as f:
        json.dump(realm, f)

    

    with open(os.path.join(config_folder, f"{cluster_config_dict['cluster_name']}.yaml"), "w") as f:
        yaml.dump(cluster_config_dict, f)

if __name__ == "__main__":
    main()


import base64
import json
import os
import subprocess
from pathlib import Path
from pprint import pprint
from secrets import token_urlsafe
from typing import Dict
from omegaconf import OmegaConf
import kubernetes
import toml
import yaml
from MAIA.helm_values import read_config_dict_and_generate_helm_values_dict
from kubernetes import client, config
from kubernetes.client.rest import ApiException


def create_config_map_from_data(data: str, config_map_name: str, namespace: str, kubeconfig_dict: Dict,
                                data_key: str = "values.yaml"):
    """
    Function to create a ConfigMap on a Kubernetes Cluster.

    Parameters
    ----------
    data :
        String containing the content of the ConfigMap to dump.
    config_map_name :
        ConfigMap name.
    namespace   :
        Namespace where to create the ConfigMap
    data_key    :
        value to use as the filename for the content in the ConfigMap.
    kubeconfig_dict :
        Kube Configuration dictionary for Kubernetes cluster authentication.
    """
    config.load_kube_config_from_dict(kubeconfig_dict)
    metadata = kubernetes.client.V1ObjectMeta(

        name=config_map_name,
        namespace=namespace,
    )

    if type(data_key)==list and type(data)==list:
        configmap = kubernetes.client.V1ConfigMap(
            api_version="v1",
            kind="ConfigMap",
            data={data_key[i]: data[i] for i in range(len(data))},
            metadata=metadata
        )
    else:
        configmap = kubernetes.client.V1ConfigMap(
            api_version="v1",
            kind="ConfigMap",
            data={data_key: data},
            metadata=metadata
        )

    with kubernetes.client.ApiClient() as api_client:
        api_instance = kubernetes.client.CoreV1Api(api_client)

        pretty = 'true'
        try:
            api_response = api_instance.create_namespaced_config_map(namespace, configmap, pretty=pretty)
            pprint(api_response)
        except ApiException as e:
            print("Exception when calling CoreV1Api->delete_namespaced_config_map: %s\n" % e)


def get_ssh_ports(n_requested_ports, port_type, ip_range, maia_metallb_ip=None):
    """
    Retrieve a list of available SSH ports based on the specified criteria.

    Args:
        n_requested_ports (int): The number of SSH ports requested.
        port_type (str): The type of port to search for ('LoadBalancer' or 'NodePort').
        ip_range (tuple): A tuple specifying the range of IPs to search within (start, end).
        maia_metallb_ip (str, optional): The specific IP address to match for 'LoadBalancer' type. Defaults to None.

    Returns:
        list: A list of available SSH ports that meet the specified criteria.
        None: If an error occurs during the process.
    """

    kubeconfig = yaml.safe_load(Path(os.environ["KUBECONFIG"]).read_text())
    config.load_kube_config_from_dict(kubeconfig)

    v1 = client.CoreV1Api()

    try:
        used_port = []
        services = v1.list_service_for_all_namespaces(watch=False)
        for svc in services.items:
            if port_type == 'LoadBalancer':
                if svc.status.load_balancer.ingress is not None:

                    if svc.spec.type == 'LoadBalancer' and svc.status.load_balancer.ingress[0].ip == maia_metallb_ip:
                        for port in svc.spec.ports:
                            if port.name == 'ssh': # TODO: or port.name == 'pt-orthanc'
                                used_port.append(int(port.port))
            elif port_type == "NodePort":
                if svc.spec.type == 'NodePort':
                    for port in svc.spec.ports:
                        used_port.append(int(port.port))

        ports = []

        for request in range(n_requested_ports):
            for port in range(ip_range[0], ip_range[1]):
                if port not in used_port:
                    ports.append(port)
                    used_port.append(port)
                    break

        return ports
    except:
        return None


def convert_username_to_jupyterhub_username(username):
    return username.replace("-", "-2d").replace("@", "-40").replace(".", "-2e")

def encode_docker_registry_secret(docker_server, docker_username, docker_password):
    auth = base64.b64encode(f"{docker_username}:{docker_password}".encode("utf-8")).decode("utf-8")
    return base64.b64encode(json.dumps({
                        "auths": {
                            docker_server: {
                                "username": docker_username,
                                "password": docker_password,
                                "auth": auth,
                            }
                        }
                    }).encode("utf-8")).decode("utf-8")


def deploy_oauth2_proxy(cluster_config, user_config, config_folder=None):
    """
    Deploys an OAuth2 Proxy using the provided cluster and user configurations.
    Args:
        cluster_config (dict): Configuration dictionary for the cluster. Expected keys include:
            - "keycloak": A dictionary with "issuer_url", "client_id", and "client_secret".
            - "domain": The domain name for the cluster.
            - "url_type": The type of URL, either "subpath" or other.
            - "storage_class": The storage class for Redis.
            - "nginx_cluster_issuer" (optional): The cluster issuer for NGINX.
            - "traefik_resolver" (optional): The resolver for Traefik.
        user_config (dict): Configuration dictionary for the user. Expected keys include:
            - "group_ID": The group ID for the user.
            - "group_subdomain": The subdomain for the user's group.
        config_folder (str, optional): The folder path where the configuration files will be saved. Defaults to None.
    Returns:
        dict: A dictionary containing deployment details:
            - "namespace": The namespace for the deployment.
            - "release": The release name for the deployment.
            - "chart": The chart name for the deployment.
            - "repo": The repository URL for the chart.
            - "version": The chart version.
            - "values": The path to the generated values YAML file.
    """

    config_file = {
        "oidc_issuer_url": cluster_config["keycloak"]["issuer_url"],
        "provider": "oidc",
        "upstreams": ["static://202"],
        "http_address": "0.0.0.0:4180",
        "oidc_groups_claim": "groups",
        "skip_jwt_bearer_tokens": True,
        "oidc_email_claim": "email",
        "allowed_groups": ["MAIA:"+user_config["group_ID"],"MAIA:admin"],
        "scope": "openid email profile",
        "redirect_url": "https://{}.{}/oauth2/callback".format(user_config["group_subdomain"],cluster_config["domain"]),
        "email_domains": ["*"],
        "proxy_prefix": "/oauth2",
        "ssl_insecure_skip_verify": True,
        "insecure_oidc_skip_issuer_verification": True,
        "cookie_secure": True,
        "reverse_proxy": True,
        "pass_access_token": True,
        "pass_authorization_header": True,
        "set_authorization_header": True,
        "set_xauthrequest": True,
        "pass_user_headers": True,
        "whitelist_domains": ["*"]
    }
    #proxy_prefix = "/oauth2-demo"
    #upstreams = [ "file:///dev/null" ]


    if cluster_config["url_type"] == "subpath":
        config_file["redirect_url"] = "https://{}/oauth2-{}/callback".format(cluster_config["domain"],user_config["group_subdomain"])
        config_file["proxy_prefix"] = "/oauth2-{}".format(user_config["group_subdomain"])

    oauth2_proxy_config = {
        "config": {
            "clientID": cluster_config["keycloak"]["client_id"],
            "clientSecret": cluster_config["keycloak"]["client_secret"],
            "cookieSecret": token_urlsafe(16),
            "configFile": toml.dumps(config_file)

        },
        "redis":{
            "enabled": True,
            "global": {
                "storageClass": cluster_config["storage_class"]
            }

        },
        "sessionStorage": {
            "type": "redis",
        },
        "image": {
            "repository": "quay.io/oauth2-proxy/oauth2-proxy",
            "tag": "",
            "pullPolicy": "IfNotPresent"
        },
        "service": {
            "type": "ClusterIP",
            "portNumber": 80,
            "appProtocol": "https",
            "annotations": {

            }
        },
        "serviceAccount": {
            "enabled": True,
            "name": "",
            "automountServiceAccountToken": True,
            "annotations": {

            }},
        "ingress": {
            "enabled": True,
            "path": "/oauth2",
            "pathType": "Prefix",
            "tls":[{"secretName":"{}.{}-tls".format(user_config["group_subdomain"],cluster_config["domain"]),"hosts": ["{}.{}".format(user_config["group_subdomain"],cluster_config["domain"])]}],
            "hosts": ["{}.{}".format(user_config["group_subdomain"],cluster_config["domain"])],
            "annotations": {

            }
        }
    }
    if cluster_config["url_type"] == "subpath":
        oauth2_proxy_config["ingress"]["hosts"] = [cluster_config["domain"]]
        oauth2_proxy_config["ingress"]["tls"][0]["hosts"] = [cluster_config["domain"]]
        oauth2_proxy_config["ingress"]["path"] = "/oauth2-{}".format(user_config["group_subdomain"])
    if "nginx_cluster_issuer" in cluster_config:
        oauth2_proxy_config["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = cluster_config["nginx_cluster_issuer"]
    if "traefik_resolver" in cluster_config:
        oauth2_proxy_config["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        oauth2_proxy_config["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = 'true'
        oauth2_proxy_config["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = cluster_config["traefik_resolver"]


    oauth2_proxy_config["chart_name"] = "oauth2-proxy"
    oauth2_proxy_config["chart_version"] = "7.7.8"
    oauth2_proxy_config["repo_url"] = "https://oauth2-proxy.github.io/manifests"  # TODO: Change this to updated values
     
    Path(config_folder).joinpath(user_config["group_ID"],"oauth2_proxy_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(user_config["group_ID"],"oauth2_proxy_values","oauth2_proxy_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(oauth2_proxy_config))

    return {
        "namespace": user_config["group_ID"].lower().replace("_", "-"),
        "release": user_config["group_ID"].lower().replace("_", "-")+"-oauth2-proxy",
        "chart": oauth2_proxy_config["chart_name"],
        "repo": oauth2_proxy_config["repo_url"],
        "version": oauth2_proxy_config["chart_version"],
        "values": str(Path(config_folder).joinpath(user_config["group_ID"],"oauth2_proxy_values","oauth2_proxy_values.yaml"))
    }
    

def deploy_mysql(cluster_config, user_config, config_folder, mysql_configs ):
    """
    Deploys a MySQL instance on a Kubernetes cluster using Helm.
    Args:
        cluster_config (dict): Configuration dictionary for the cluster, including storage class.
        user_config (dict): Configuration dictionary for the user, including group ID.
        config_folder (str): Path to the folder where configuration files will be stored.
        mysql_configs (dict): Configuration dictionary for MySQL, including user, password, and other settings.
    Returns:
        dict: A dictionary containing deployment details such as namespace, release name, chart name, repository URL, version, and values file path.
    """

    namespace = user_config["group_ID"].lower().replace("_", "-")
    kubeconfig = yaml.safe_load(Path(os.environ["KUBECONFIG"]).read_text())

    
    mysql_config = {
        "namespace": namespace,
        "chart_name": "mysql-db-v1",
        "docker_image": "mysql",
        "tag": "8.0.28",
        "memory_request": "2Gi",
        "cpu_request": "500m",
        "deployment": True,
        "ports": {
            "mysql": [
                3306
            ]
        },
        "persistent_volume": [
            {
                "mountPath": "/var/lib/mysql",
                "size": "20Gi",
                "access_mode": "ReadWriteMany",
                "pvc_type": cluster_config["storage_class"]
            }
        ],
        "env_variables": {
            "MYSQL_ROOT_PASSWORD": mysql_configs.get("mysql_password","root") ,
            "MYSQL_USER": mysql_configs.get("mysql_user","root"),
            "MYSQL_PASSWORD": mysql_configs.get("mysql_password","root"),
            "MYSQL_DATABASE": "mysql"
        }
    } # TODO: Change this to updated values


    mysql_values = read_config_dict_and_generate_helm_values_dict(mysql_config, kubeconfig)

    mysql_values["chart_name"] = "mkg"
    mysql_values["chart_version"] = "1.0.2"
    mysql_values["repo_url"] = "https://kthcloud.github.io/MAIA/" # TODO: Change this to updated values
     
    Path(config_folder).joinpath(user_config["group_ID"],"mysql_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(user_config["group_ID"],"mysql_values","mysql_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(mysql_values))

    return {
        "namespace": user_config["group_ID"].lower().replace("_", "-"),
        "release": user_config["group_ID"].lower().replace("_", "-")+"-mysql",
        "chart": mysql_values["chart_name"],
        "repo": mysql_values["repo_url"],
        "version": mysql_values["chart_version"],
        "values": str(Path(config_folder).joinpath(user_config["group_ID"],"mysql_values","mysql_values.yaml"))
    }


def deploy_mlflow(cluster_config, user_config, config_folder, mysql_config=None, minio_config=None):
    """
    Deploys an MLflow instance on a Kubernetes cluster using Helm.
    Args:
        cluster_config (dict): Configuration dictionary for the Kubernetes cluster.
        user_config (dict): Configuration dictionary for the user, including group_ID.
        config_folder (str): Path to the folder where configuration files will be stored.
        mysql_config (dict, optional): Configuration dictionary for MySQL, including mysql_user and mysql_password. Defaults to None.
        minio_config (dict, optional): Configuration dictionary for MinIO, including console_access_key and console_secret_key. Defaults to None.
    Returns:
        dict: A dictionary containing deployment details such as namespace, release name, chart name, repository URL, chart version, and path to the values file.
    """
    
    namespace = user_config["group_ID"].lower().replace("_", "-")

    kubeconfig = yaml.safe_load(Path(os.environ["KUBECONFIG"]).read_text())
    config.load_kube_config_from_dict(kubeconfig)

    mlflow_config = {
    "namespace": namespace,
    "chart_name": "mlflow-v1",
        "docker_image": "kthcloud/mlflow", #TODO: Make this configurable
        "tag": "1.1",
    "memory_request": "2Gi",
    "cpu_request": "500m",
    "allocationTime": "180d",
        "ports": {
        "mlflow": [
            5000
        ]
        },
    "user_secret": [
        namespace
    ],
    "user_secret_params": [
        "user",
        "password"
    ],
    
    "env_variables": {
        "MYSQL_URL": "{}-mysql-mkg".format(namespace),
        "MYSQL_PASSWORD": mysql_config.get("mysql_password","root"),
        "MYSQL_USER": mysql_config.get("mysql_user","root"),
        "BUCKET_NAME": "mlflow",
        "BUCKET_PATH": "mlflow",
        "AWS_ACCESS_KEY_ID": minio_config.get("console_access_key","minio"),
        "AWS_SECRET_ACCESS_KEY": minio_config.get("console_secret_key","minio"),
        "MLFLOW_S3_ENDPOINT_URL": "http://minio:80"
    }
    } # TODO: Change this to updated values
    if "imagePullSecrets" in cluster_config:
        mlflow_config["image_pull_secret"] = cluster_config["imagePullSecrets"]
    
    mlflow_values = read_config_dict_and_generate_helm_values_dict(mlflow_config, kubeconfig)

    mlflow_values["chart_name"] = "mkg"
    mlflow_values["chart_version"] = "1.0.2"
    mlflow_values["repo_url"] = "https://kthcloud.github.io/MAIA/"  # TODO: Change this to updated values
     
    Path(config_folder).joinpath(user_config["group_ID"],"mlflow_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(user_config["group_ID"],"mlflow_values","mlflow_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(mlflow_values))

    return {
        "namespace": user_config["group_ID"].lower().replace("_", "-"),
        "release": user_config["group_ID"].lower().replace("_", "-")+"-mlflow",
        "chart": mlflow_values["chart_name"],
        "repo": mlflow_values["repo_url"],
        "version": mlflow_values["chart_version"],
        "values": str(Path(config_folder).joinpath(user_config["group_ID"],"mlflow_values","mlflow_values.yaml"))
    }


#!/usr/bin/env python

import json
from pathlib import Path

import click
import yaml

import datetime
import json

import MAIA
from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from textwrap import dedent
from MAIA.maia_fn import convert_username_to_jupyterhub_username
version = MAIA.__version__

TIMESTAMP = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.datetime.now())
from MAIA.maia_fn import get_ssh_ports


DESC = dedent(
    """
    Script to deploy the MAIA-Addons helm chart to a Kubernetes cluster. The target cluster is specified by setting the corresponding ``--cluster--config-file``,
    while the namespace-related configuration is specified with ``--form``.
    """  # noqa: E501
)
EPILOG = dedent(
    """
    Example call:
    ::
        {filename}  --form /PATH/TO/form.yaml --cluster-config-file /PATH/TO/cluster.yaml
    """.format(  # noqa: E501
        filename=Path(__file__).stem
    )
)

def get_arg_parser():
    pars = ArgumentParser(description=DESC, epilog=EPILOG, formatter_class=RawTextHelpFormatter)

    pars.add_argument(
        "--form",
        type=str,
        required=True,
        help="YAML configuration file used to extract the namespace configuration.",
    )
    pars.add_argument(
        "--maia-config-file",
        type=str,
        required=True,
        help="YAML configuration file used to extract the MAIA configuration.",
    )
    pars.add_argument(
        "--cluster-config-file",
        type=str,
        required=True,
        help="YAML configuration file used to extract the cluster configuration.",
    )

    pars.add_argument('-v', '--version', action='version', version='%(prog)s ' + version)

    return pars

@click.command()
@click.option("--form", type=str)
@click.option("--maia-config-file", type=str)
@click.option("--cluster-config-file", type=str)
def create_maia_addons_config(form,maia_config_file,cluster_config_file):
    create_maia_addons_config_api(form,maia_config_file,cluster_config_file)

def create_maia_addons_config_api( form,
                                  maia_config_file,
                              cluster_config_file,
                                   config_folder = None
                             ):

    with open(maia_config_file, "r") as f:
        maia_form = yaml.safe_load(f)

    with open(cluster_config_file,"r") as f:
        cluster_config = yaml.safe_load(f)


    with open(form,"r") as f:
        user_form = yaml.safe_load(f)

    group_subdomain = user_form["group_subdomain"]
    namespace = user_form["group_ID"].lower().replace("_", "-")


    users = user_form["users"]

    jupyterhub_users = []

    oauth_url = user_form["oauth_url"]

    for user in users:
        jupyterhub_users.append(convert_username_to_jupyterhub_username(user))

    domain = cluster_config["domain"]

    if "url_type" in cluster_config:
        if cluster_config["url_type"] == "subdomain":
            hub_address = f"{group_subdomain}.{domain}"
        elif cluster_config["url_type"] == "subpath":
            hub_address = domain
        else:
            hub_address = None

    maia_addons_values_template = {
        "namespace": namespace.lower(),
        "oauth_url": oauth_url,
        "ingress_host": hub_address,
        "metallb": {

        },
        "ssh_port_type": cluster_config["ssh_port_type"],
        "users": []
    }
    if "metallb_shared_ip" in cluster_config:
        maia_addons_values_template["metallb"]["shared_ip"] = cluster_config["metallb_shared_ip"]
    if "metallb_ip_pool" in cluster_config:
        maia_addons_values_template["metallb"]["ip_pool"] = cluster_config["metallb_ip_pool"]

    metallb_shared_ip = cluster_config["maia_metallb_ip"] if "maia_metallb_ip" in cluster_config else None
    maia_addons_values_template["metallb"]["load_balancer_ip"] = metallb_shared_ip
    ssh_ports = get_ssh_ports(len(users),cluster_config["ssh_port_type"],cluster_config["port_range"],maia_metallb_ip=metallb_shared_ip)
    for user, jupyterhub_user,ssh_port in zip(users,jupyterhub_users, ssh_ports):
        maia_addons_values_template["users"].append(
            {"jupyterhub_username": jupyterhub_user,
             "username": user,
             "ssh_port": ssh_port
             })


    if "traefik_resolver" in cluster_config:
        maia_addons_values_template["traefik"] = {
            "enabled": True,
            "traefik_resolver": cluster_config["traefik_resolver"]
        }


    if "nginx_cluster_issuer" in cluster_config:
        maia_addons_values_template["nginx"] = {
            "enabled": True,
            "cluster_issuer": cluster_config["nginx_cluster_issuer"]
        }

    if "minio_console_service" in user_form:
       
        cluster_config["nginx_proxy_image"] = maia_form["nginx_proxy_image"]
        maia_addons_values_template["proxy_nginx"] = {
            "enabled": True,
            "imagePullSecrets": cluster_config["imagePullSecrets"],
            "image": cluster_config["nginx_proxy_image"],
            "console_service": user_form["minio_console_service"],
            #"cluster_issuer": cluster_config["nginx_cluster_issuer"],
            "mlflow": user_form["mlflow_service"],
            "console_service_path": "minio-console",
            "mlflow_path": "mlflow",
            "label_studio_path": "label-studio",
            "kubeflow_path": "kubeflow",
            "minio_path": "minio"
        }
        if cluster_config["url_type"] == "subpath":
            maia_addons_values_template["proxy_nginx"]["console_service_path"] = f"{namespace}-minio-console"
            maia_addons_values_template["proxy_nginx"]["mlflow_path"] = f"{namespace}-mlflow"
            maia_addons_values_template["proxy_nginx"]["label_studio_path"] = f"{namespace}-label-studio"
            maia_addons_values_template["proxy_nginx"]["kubeflow_path"] = f"{namespace}-kubeflow"
            maia_addons_values_template["proxy_nginx"]["minio_path"] = f"{namespace}-minio"

    maia_addons_template = {
        "resource": {
            "helm_release": {
                "maia-addons": {
                    "name": "maia-addons-{}".format(namespace.lower()),
                    "repository": "https://kthcloud.github.io/MAIA/",
                    "chart": "maia-addons",
                    "version": maia_form["maia_addons_version"],
                    "namespace": namespace.lower(),
                    "create_namespace": False,
                    "values": [
                        yaml.dump(maia_addons_values_template)
                    ]

                }
            }
        }
    }

    if config_folder is None:
        config_folder = "."
    
    group_id = user_form["group_ID"]
    Path(config_folder).joinpath(user_form["group_ID"]).mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(user_form["group_ID"],f"{group_id}_maia_addons.tf.json"), "w") as f:
        json.dump(maia_addons_template, f, indent=2)

    with open(Path(config_folder).joinpath(user_form["group_ID"],f"{group_id}_maia_addons_values.yaml"), "w") as f:
        print(maia_addons_template["resource"]["helm_release"]["maia-addons"]["values"][0], file=f)

    helm_namespace = maia_addons_template["resource"]["helm_release"]["maia-addons"]["namespace"]
    helm_name = maia_addons_template["resource"]["helm_release"]["maia-addons"]["name"]
    helm_chart = maia_addons_template["resource"]["helm_release"]["maia-addons"]["chart"]
    helm_repo = maia_addons_template["resource"]["helm_release"]["maia-addons"]["repository"]
    helm_repo_version = maia_addons_template["resource"]["helm_release"]["maia-addons"]["version"]

    config_path = Path(config_folder).joinpath(user_form["group_ID"],f"{group_id}_maia_addons_values.yaml")
    cmds =[
        # "Run the following command to deploy JupyterHub: ",
        f"helm repo add maia {helm_repo}",
        f"helm repo update",
        f"helm upgrade --install -n {helm_namespace} {helm_name} maia/{helm_chart} --values {config_path} --version={helm_repo_version}"

    ]
    print("\n".join(cmds))

    return cmds


def main():
    create_maia_addons_config()


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Dict, Any




def read_config_dict_and_generate_helm_values_dict(config_dict: Dict[str, Any], kubeconfig_dict: Dict[str, Any]) -> \
        Dict[str, Any]:
    """
    Read Config Dict and generate helm-ready values dict.

    Parameters
    ----------
    config_dict :
        Config Dict
    kubeconfig_dict :
        Kubeconfig Dict

    Returns
    -------
    Helm values dict

    """
    from MAIA.maia_fn import create_config_map_from_data
    value_dict = {}
    value_dict["image"] = {}
    value_dict["image"]["repository"] = config_dict['docker_image']

    if 'tag' not in config_dict:
        value_dict["image"]["tag"] = "latest"

    else:
        value_dict["image"]["tag"] = config_dict['tag']

    value_dict["extraVolumeMounts"] = [{"name": "dshm", "mountPath": "/dev/shm"}]

    if 'memory_request' in config_dict:
        value_dict["resources"] = {}
        value_dict["resources"]["requests"] = {}
        value_dict["resources"]["limits"] = {}
        value_dict["resources"]["requests"]["memory"] = config_dict['memory_request']
        value_dict["resources"]["limits"]["memory"] = config_dict['memory_request']

    if 'cpu_request' in config_dict:
        if "resources" not in value_dict:
            value_dict["resources"] = {}
            value_dict["resources"]["requests"] = {}
            value_dict["resources"]["limits"] = {}

        value_dict["resources"]["requests"]["cpu"] = config_dict['cpu_request']
        value_dict["resources"]["limits"]["cpu"] = config_dict['cpu_request']

    if "allocationTime" in config_dict:
        time_multiplier = 24 * 60 * 60  # Days
        if config_dict["allocationTime"][-1] == "s":
            time_multiplier = 1
            config_dict["allocationTime"] = config_dict["allocationTime"][:-1]
        elif config_dict["allocationTime"][-1] == "m":
            time_multiplier = 60
            config_dict["allocationTime"] = config_dict["allocationTime"][:-1]
        elif config_dict["allocationTime"][-1] == "h":
            time_multiplier = 24 * 60
            config_dict["allocationTime"] = config_dict["allocationTime"][:-1]
        elif config_dict["allocationTime"][-1] == "d":
            time_multiplier = 24 * 60 * 60
            config_dict["allocationTime"] = config_dict["allocationTime"][:-1]
    else:
        time_multiplier = 1
        config_dict["allocationTime"] = 0

    value_dict["allocatedTime"] = int(config_dict["allocationTime"]) * time_multiplier

    if 'gpu_request' in config_dict and int(config_dict["gpu_request"]) > 0:
        if "resources" not in value_dict:
            value_dict["resources"] = {}
            value_dict["resources"]["limits"] = {}
        value_dict["resources"]["limits"]["nvidia.com/gpu"] = config_dict["gpu_request"]
        value_dict["gpus"] = config_dict["gpu_request"]

    if 'persistent_volume' in config_dict:
        for idx, persistent_volume in enumerate(config_dict['persistent_volume']):
            if idx == 0:
                value_dict["persistentVolume"] = [{}]
            else:
                value_dict["persistentVolume"].append({})
            for persistent_volume_key in persistent_volume:
                value_dict["persistentVolume"][-1][persistent_volume_key] = config_dict['persistent_volume'][idx][
                    persistent_volume_key]

    if 'existing_persistent_volume' in config_dict:
        for idx, persistent_volume in enumerate(config_dict['existing_persistent_volume']):
            if idx == 0:
                value_dict["existingPersistentVolume"] = [{}]
            else:
                value_dict["existingPersistentVolume"].append({})
            for persistent_volume_key in persistent_volume:
                value_dict["existingPersistentVolume"][-1][persistent_volume_key] = \
                    config_dict['existing_persistent_volume'][idx][
                        persistent_volume_key]

    if 'user_secret' in config_dict:
        value_dict["extraEnv"] = [{"name": "n_users", "value": str(len(config_dict['user_secret']))}]
        value_dict["user_secrets"] = config_dict["user_secret"]
        if len(config_dict['user_secret']) == 1:
            for user_secret_param in config_dict['user_secret_params']:
                value_dict["extraEnv"].append({
                    "name": user_secret_param,
                    "valueFrom": {"secretKeyRef": {"key": user_secret_param, "name": config_dict['user_secret'][0]}},

                })
            # value_dict["secretCSIVolume"] = [{"name":"{}-vault-secret".format(config_dict['user_secret'][0].split("-")[1])}]
            # value_dict["extraVolumeMounts"].append({
            #    "name": "csi-secret-0",
            #    "mountPath": "/mnt/secrets-store-0"
            #    #"readOnly": "true"
            # })

        else:
            for user_id, user in enumerate(config_dict['user_secret']):
                for idx, user_secret_param in enumerate(config_dict['user_secret_params']):
                    value_dict["extraEnv"].append({
                        "name": user_secret_param + "_{}".format(user_id),
                        "valueFrom": {
                            "secretKeyRef": {"key": user_secret_param, "name": user}},

                    })
                # if idx == 0:
                #    value_dict["secretCSIVolume"] = [{"name": user}]
                # else:
                #    value_dict["secretCSIVolume"].append({"name": "{}-vault-secret".format(user.split("-")[1])})
                # value_dict["extraVolumeMounts"].append({
                #    "name": "csi-secret-{}".format(user_id),
                #    "mountPath": "/mnt/secrets-store-{}".format(user_id),
                #    "readOnly": "true"
                # })

    single_config_file = True
    if 'mount_files' in config_dict:
        if single_config_file:
            for idx, mount_file in enumerate(config_dict["mount_files"]):

                if idx == 0:
                    if len(config_dict['mount_files'][mount_file]) > 2 and config_dict['mount_files'][mount_file][
                        2] == "readOnly":
                        value_dict["extraVolumeMounts"].append({
                            "name": mount_file.lower().replace("_", "-"),
                            "mountPath": config_dict['mount_files'][mount_file][1],
                            "readOnly": True
                        })
                    else:
                        value_dict["extraVolumeMounts"].append({
                            "name": mount_file.lower().replace("_", "-"),
                            "mountPath": config_dict['mount_files'][mount_file][1],
                            "readOnly": False
                        })

                    value_dict["extraConfigMapVolumes"] = []

                    value_dict["extraConfigMapVolumes"].append({
                        "name": mount_file.lower().replace("_", "-"),
                        "configMapName": mount_file.lower().replace("_", "-"),
                        "configMapFile": Path(config_dict['mount_files'][mount_file][0]).name,
                        "configMapPath": Path(config_dict['mount_files'][mount_file][0]).name
                    })

            files = []
            file_names = []
            mount_file_config = ""
            for idx, mount_file in enumerate(config_dict["mount_files"]):
                if idx == 0:
                    mount_file_config = mount_file
                with open(config_dict['mount_files'][mount_file][0], "r") as f:
                    files.append(f.read())
                    file_names.append(Path(config_dict['mount_files'][mount_file][0]).name)
                    # d = json.load(f)
            create_config_map_from_data(  # json.dumps(d),
                files,
                mount_file_config.lower().replace("_", "-"),
                config_dict["namespace"],
                kubeconfig_dict,
                file_names)




        else:
            for idx, mount_file in enumerate(config_dict["mount_files"]):

                if len(config_dict['mount_files'][mount_file]) > 2 and config_dict['mount_files'][mount_file][
                    2] == "readOnly":
                    value_dict["extraVolumeMounts"].append({
                        "name": mount_file.lower().replace("_", "-"),
                        "mountPath": config_dict['mount_files'][mount_file][1],
                        "readOnly": True
                    })
                else:
                    value_dict["extraVolumeMounts"].append({
                        "name": mount_file.lower().replace("_", "-"),
                        "mountPath": config_dict['mount_files'][mount_file][1],
                        "readOnly": False
                    })

                with open(config_dict['mount_files'][mount_file][0], "r") as f:
                    #d = json.load(f)
                    create_config_map_from_data(#json.dumps(d),
                                                f.read(),
                                                mount_file.lower().replace("_", "-"),
                                                config_dict["namespace"],
                                                kubeconfig_dict,
                                                Path(config_dict['mount_files'][mount_file][0]).name)

                if idx == 0:
                    value_dict["extraConfigMapVolumes"] = []

                value_dict["extraConfigMapVolumes"].append({
                    "name": mount_file.lower().replace("_", "-"),
                    "configMapName": mount_file.lower().replace("_", "-"),
                    "configMapFile": Path(config_dict['mount_files'][mount_file][0]).name,
                    "configMapPath": Path(config_dict['mount_files'][mount_file][0]).name
                })

    if 'env_variables' in config_dict:
        if "extraEnv" not in value_dict:
            value_dict["extraEnv"] = []
        for env_variable in config_dict["env_variables"]:
            value_dict["extraEnv"].append({
                "name": env_variable,
                "value": config_dict['env_variables'][env_variable]

            })

    if 'ports' in config_dict:
        value_dict["serviceEnabled"] = "True"

        if 'service_type' in config_dict:
            value_dict["service"] = {}
            value_dict["service"]["type"] = config_dict['service_type']

        else:
            value_dict["service"] = {}
            value_dict["service"]["type"] = "ClusterIP"
        for idx, port in enumerate(config_dict['ports']):
            if idx == 0:
                value_dict["service"]["ports"] = [
                    {
                        "port": config_dict['ports'][port][0],
                        "targetPort": config_dict['ports'][port][0],
                        "name": port
                    }]
                if 'service_type' in config_dict and config_dict['service_type'] != "ClusterIP":
                    value_dict["service"]["ports"][-1]["nodePort"] = config_dict['ports'][port][1]
            else:
                value_dict["service"]["ports"].append({
                    "port": config_dict['ports'][port][0],
                    "targetPort": config_dict['ports'][port][0],
                    "name": port
                })
                if 'service_type' in config_dict and config_dict['service_type'] != "ClusterIP":
                    value_dict["service"]["ports"][-1]["nodePort"] = config_dict['ports'][port][1]

    if "ingress" in config_dict:
        value_dict["ingress"] = {}

        if "traefik_resolver" in config_dict["ingress"]:
            value_dict["ingress"]["traefik"] = {"enabled": True}
            value_dict["ingress"]["traefik_resolver"] = config_dict["ingress"][
            "traefik_resolver"]
        if "traefik_middleware" in config_dict["ingress"]:
            value_dict["ingress"]["traefik"] = {"enabled": True}
            value_dict["ingress"]["traefik_middleware"] = config_dict["ingress"][
            "traefik_middleware"]

        if "nginx_issuer" in config_dict["ingress"]:
            value_dict["ingress"]["nginx"] = {"enabled": True}
            value_dict["ingress"]["nginx_issuer"] = config_dict["ingress"]["nginx_issuer"]

        value_dict["ingress"]["host"] = config_dict["ingress"]["host"]
        if "path" in config_dict["ingress"]:
            value_dict["ingress"]["path"] = config_dict["ingress"]["path"]
        value_dict["ingress"]["port"] = config_dict["ingress"]["port"]
        if "oauth_url" in config_dict["ingress"]:
            value_dict["ingress"]["oauth_url"] = config_dict["ingress"]["oauth_url"]

    if "node_selector" in config_dict:
        value_dict["nodeSelected"] = config_dict["node_selector"]

    if "gpu_selector" in config_dict:
        for gpu_selection in config_dict["gpu_selector"]:
            value_dict["gpuSelected"] = {}
            value_dict["gpuSelected"][gpu_selection] = config_dict["gpu_selector"][gpu_selection]

    if "deployment" in config_dict:
        value_dict["deploy_as_job"] = False
    if "command" in config_dict:
        value_dict["command"] = config_dict["command"]

    if "image_pull_secret" in config_dict:
        value_dict["imagePullSecrets"] = [{"name": config_dict["image_pull_secret"]}]
    return value_dict




from pathlib import Path
import click
import yaml
from omegaconf import OmegaConf
from MAIA.maia_admin import install_maia_project, create_harbor_values, create_keycloak_values, create_loginapp_values, create_minio_operator_values, create_maia_admin_toolkit_values
from hydra import initialize, initialize_config_dir
from hydra import compose as hydra_compose
import datetime

import os
import asyncio
from pyhelm3 import Client

from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from textwrap import dedent

import MAIA
version = MAIA.__version__


TIMESTAMP = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.datetime.now())

DESC = dedent(
    """
    Script to Install MAIA Admin Toolkit to a Kubernetes cluster from ArgoCD. The specific MAIA configuration is specified by setting the corresponding ``--maia-config-file``,
    and the cluster configuration is specified by setting the corresponding ``--cluster-config``.
    """  # noqa: E501
)
EPILOG = dedent(
    """
    Example call:
    ::
        {filename} --maia-config-file /PATH/TO/maia_config.yaml --cluster-config /PATH/TO/clusyer_config.yaml --config-folder /PATH/TO/config_folder
    """.format(  # noqa: E501
        filename=Path(__file__).stem
    )
)

def get_arg_parser():
    pars = ArgumentParser(description=DESC, epilog=EPILOG, formatter_class=RawTextHelpFormatter)


    pars.add_argument(
        "--maia-config-file",
        type=str,
        required=True,
        help="YAML configuration file used to extract MAIA configuration.",
    )

    pars.add_argument(
        "--cluster-config",
        type=str,
        required=True,
        help="YAML configuration file used to extract the cluster configuration.",
    )

    pars.add_argument(
        "--config-folder",
        type=str,
        required=True,
        help="Configuration Folder where to locate (and temporarily store) the MAIA configuration files.",
    )

    pars.add_argument('-v', '--version', action='version', version='%(prog)s ' + version)

    return pars


async def verify_installed_maia_admin_toolkit(project_id, namespace):

    print(os.environ["KUBECONFIG"])
    client = Client(kubeconfig = os.environ["KUBECONFIG"])


    try:
        revision = await client.get_current_revision(project_id, namespace = namespace)
    except:
        print("Project not found")
        return -1
    chart_metadata = await revision.chart_metadata()
    print(
        revision.release.name,
        revision.release.namespace,
        revision.revision,
        str(revision.status),
        chart_metadata.name,
        chart_metadata.version
    )
    return revision.revision

@click.command()
@click.option("--maia-config-file", type=str)
@click.option("--cluster-config", type=str)
@click.option("--config-folder", type=str)
def main(maia_config_file, cluster_config, config_folder):
    install_maia_admin_toolkit(maia_config_file,cluster_config, config_folder)

def install_maia_admin_toolkit(maia_config_file, cluster_config, config_folder):

    maia_config_dict = yaml.safe_load(Path(maia_config_file).read_text())

    cluster_config_dict = yaml.safe_load(Path(cluster_config).read_text())

    admin_group_ID = maia_config_dict["admin_group_ID"]
    project_id = "maia-admin"

    cluster_address = "https://kubernetes.default.svc" #TODO: Change this to make it configurable

    helm_commands = []

    helm_commands.append(create_harbor_values(config_folder, project_id, cluster_config_dict))
    helm_commands.append(create_keycloak_values(config_folder, project_id,cluster_config_dict))
    helm_commands.append(create_loginapp_values(config_folder, project_id,cluster_config_dict))
    helm_commands.append(create_minio_operator_values(config_folder, project_id,cluster_config_dict))
    helm_commands.append(create_maia_admin_toolkit_values(config_folder, project_id,cluster_config_dict, maia_config_dict=maia_config_dict))
    

    for helm_command in helm_commands:
        cmd = [
            "helm",
            "upgrade",
            "--install",
            "--wait",
            "-n",
            helm_command["namespace"],
            helm_command["release"],
            helm_command["chart"],
            "--repo",
            helm_command["repo"],
            "--version",
            helm_command["version"],
            "--values",
            helm_command["values"],
        ]
        print(" ".join(cmd))

    values = {
        "defaults": [
              "_self_",
              {"harbor_values": "harbor_values"},
                {"keycloak_values": "keycloak_values"},
                {"loginapp_values": "loginapp_values"},
                {"minio_operator_values": "minio_operator_values"},
                {"maia_admin_toolkit_values": "maia_admin_toolkit_values"},
              
         ],
        "argo_namespace": maia_config_dict["argocd_namespace"],
        "admin_group_ID": admin_group_ID,
        "destination_server": f"{cluster_address}",
        "sourceRepos": [
            "https://kthcloud.github.io/MAIA/",
            "https://helm.goharbor.io",
            "https://charts.bitnami.com/bitnami",
            "https://storage.googleapis.com/loginapp-releases/charts/",
            "https://operator.min.io"
            
        ]
    }
    Path(config_folder).joinpath(project_id).mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id,"values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(values))




    initialize_config_dir(config_dir=str(Path(config_folder).joinpath(project_id)), job_name=project_id)
    cfg = hydra_compose("values.yaml")
    OmegaConf.save(cfg, str(Path(config_folder).joinpath(project_id, f"{project_id}_values.yaml")), resolve=True)

    

    revision = asyncio.run(verify_installed_maia_admin_toolkit(project_id, maia_config_dict["argocd_namespace"]))



    if revision == -1:
        print("Installing MAIA Admin Toolkit")
    
    
        project_chart = maia_config_dict["admin_project_chart"]
        asyncio.run(install_maia_project(project_id, Path(config_folder).joinpath(project_id, f"{project_id}_values.yaml"),maia_config_dict["argocd_namespace"], project_chart))
    else:
        print("Upgrading MAIA Admin Toolkit")
    
    
        project_chart = maia_config_dict["admin_project_chart"]
        asyncio.run(install_maia_project(project_id, Path(config_folder).joinpath(project_id, f"{project_id}_values.yaml"),maia_config_dict["argocd_namespace"], project_chart))


if __name__ == "__main__":
    main()


from pathlib import Path
from omegaconf import OmegaConf
from kubernetes import client, config

def create_prometheus_values(config_folder, project_id, cluster_config_dict, maia_config_dict):
    """
    Generates Prometheus values configuration for a Kubernetes cluster and writes it to a YAML file.
    Args:
        config_folder (str): The folder where the configuration files will be stored.
        project_id (str): The project identifier.
        cluster_config_dict (dict): Dictionary containing cluster configuration details.
        maia_config_dict (dict): Dictionary containing MAIA configuration details.
    Returns:
        dict: A dictionary containing the namespace, repository URL, chart version, path to the values file, release name, and chart name.
    """

    config.load_kube_config() 

    # Create a CoreV1Api instance
    v1 = client.CoreV1Api()

    # Get the list of nodes
    nodes = v1.list_node()

    # Extract InternalIP for each node
    internal_ips = []
    for node in nodes.items:
        for address in node.status.addresses:
            if address.type == "InternalIP":
                internal_ips.append(address.address)

    prometheus_values = {
        "namespace": "observability",
        "chart_version": "45.5.0",
        "repo_url": "https://prometheus-community.github.io/helm-charts",
        "chart_name": "kube-prometheus-stack",
    } # TODO: Change this to updated values
    
    admin_group_ID = maia_config_dict["admin_group_ID"]
    domain = cluster_config_dict["domain"]
    prometheus_values.update({
        "kubeControllerManager": {
            "endpoints": internal_ips
        },
        "kubeScheduler": {
            "endpoints": internal_ips
        },
        "grafana": {

            "grafana.ini": {
                "server": {
                    "root_url": "https://grafana."+cluster_config_dict["domain"],
                    "auth.generic_oauth": {
                        "api_url": f"https://iam.{domain}/realms/maia/protocol/openid-connect/userinfo",
                        "auth_url": f"https://iam.{domain}/realms/maia/protocol/openid-connect/auth",
                        "client_id": "maia",
                        "client_secret": cluster_config_dict["keycloak_maia_client_secret"],
                        "enabled": True,
                        "name": "OAuth",
                        "empty_scopes": False,
                        "role_attribute_path": f"contains(groups[*], '{admin_group_ID}') && 'Admin' || 'Viewer'",
                        "scopes": "openid profile email",
                        "team_ids": admin_group_ID,
                        "team_ids_attribute_path": "groups[*]",
                        "teams_url": f"https://iam.{domain}/realms/maia/protocol/openid-connect/userinfo",
                        "token_url": f"https://iam.{domain}/realms/maia/protocol/openid-connect/token"

                }
            }
            },
            "assertNoLeakedSecrets": False,
            "defaultDashboardsEnabled": True,
            "persistence": {
                "enabled": True,
            },
            "ingress": {
                "enabled": True,
                "annotations": {
                },
                "hosts": ["grafana."+cluster_config_dict["domain"]],
                "tls": [
                    {
                        
                        "hosts": ["grafana."+cluster_config_dict["domain"]]
                    }
                ]
            },
            "additionalDataSources": [
                {
                    "name": "loki",
                    "type": "loki",
                    "url": "http://loki-maia-core.observability.svc.cluster.local:3100"
                },
                {
                    "name": "tempo",
                    "type": "tempo",
                    "url": "http://tempo-maia-core.observability.svc.cluster.local:3100"
                }
            ]
        }
    })

    gpu_enabled = True


    if gpu_enabled:
        prometheus_values["prometheus"] = {
            "prometheusSpec": {
                "additionalScrapeConfigs": [
                    {
                        "job_name": "gpu-metrics",
                        "kubernetes_sd_configs": [
                            {
                                "namespaces": {
                                    "names": ["gpu-operator"]
                                },
                                "role": "endpoints"
                            }
                        ],
                        "metrics_path": "/metrics",
                        "relabel_configs": [
                            {
                                "action": "replace",
                                "source_labels": ["__meta_kubernetes_pod_node_name"],
                                "target_label": "kubernetes_node"
                            }
                        ],
                        "scheme": "http",
                        "scrape_interval": "1s"
                    }
                ],
                "serviceMonitorSelectorNilUsesHelmValues": False
            }
        }

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        prometheus_values["grafana"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        prometheus_values["grafana"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = 'true'
        prometheus_values["grafana"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = cluster_config_dict["traefik_resolver"]
    elif cluster_config_dict["ingress_class"] == "nginx":
        prometheus_values["grafana"]["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"
        prometheus_values["grafana"]["ingress"]["tls"][0]["secretName"] = "grafana."+cluster_config_dict["domain"]

    Path(config_folder).joinpath(project_id,"prometheus_values").mkdir(parents=True, exist_ok=True)
    with open( Path(config_folder).joinpath(project_id,"prometheus_values","prometheus_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(prometheus_values))

    return {
        "namespace": prometheus_values["namespace"],
        "repo": prometheus_values["repo_url"],
        "version": prometheus_values["chart_version"],
        "values": str( Path(config_folder).joinpath(project_id,"prometheus_values","prometheus_values.yaml")),
        "release": f"{project_id}-prometheus",
        "chart": prometheus_values["chart_name"]
    }

def create_loki_values(config_folder, project_id):
    """
    Creates and writes Loki values configuration to a YAML file and returns deployment details.
    Args:
        config_folder (str): The path to the configuration folder.
        project_id (str): The project identifier.
    Returns:
        dict: A dictionary containing deployment details including namespace, repo URL, chart version, 
              values file path, release name, and chart name.
    """


    loki_values = {
        "namespace": "observability",
        "chart_version": "2.9.9",
        "repo_url": "https://grafana.github.io/helm-charts",
        "chart_name": "loki-stack"
    } # TODO: Change this to updated values

    loki_values.update( {   "grafana": {
            "sidecar": {
                "datasources": {
                    "enabled": False
                }
            }
    }
    })
    

    Path(config_folder).joinpath(project_id,"loki_values").mkdir(parents=True, exist_ok=True)
    with open( Path(config_folder).joinpath(project_id,"loki_values","loki_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(loki_values))

    return {
        "namespace": loki_values["namespace"],
        "repo": loki_values["repo_url"],
        "version": loki_values["chart_version"],
        "values": str( Path(config_folder).joinpath(project_id,"loki_values","loki_values.yaml")),
        "release": f"{project_id}-loki",
        "chart": loki_values["chart_name"]
    }

def create_tempo_values(config_folder, project_id):
    """
    Creates a set of tempo values and writes them to a YAML file in the specified configuration folder.
    Args:
        config_folder (str): The path to the configuration folder where the tempo values will be stored.
        project_id (str): The project identifier used to create a subdirectory and name the YAML file.
    Returns:
        dict: A dictionary containing the following keys:
            - namespace (str): The namespace for the tempo values.
            - repo (str): The repository URL for the Helm chart.
            - version (str): The version of the Helm chart.
            - values (str): The path to the generated tempo values YAML file.
            - release (str): The release name for the Helm chart.
            - chart (str): The name of the Helm chart.
    """


    tempo_values = {
        "namespace": "observability",
        "chart_version": "1.0.0",
        "repo_url": "https://grafana.github.io/helm-charts",
        "chart_name": "tempo"
    } # TODO: Change this to updated values
    

    Path(config_folder).joinpath(project_id,"tempo_values").mkdir(parents=True, exist_ok=True)
    with open( Path(config_folder).joinpath(project_id,"tempo_values","tempo_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(tempo_values))

    return {"namespace": tempo_values["namespace"],
     "repo": tempo_values["repo_url"],
     "version": tempo_values["chart_version"],
     "values": str( Path(config_folder).joinpath(project_id,"tempo_values","tempo_values.yaml")),
     "release": f"{project_id}-tempo",
     "chart": tempo_values["chart_name"]
     }

def create_core_toolkit_values(config_folder, project_id, cluster_config_dict):
    """
    Creates and saves the core toolkit values for a Kubernetes cluster.
    This function generates a dictionary of core toolkit values based on the provided
    configuration folder, project ID, and cluster configuration dictionary. It retrieves
    the internal IP addresses of the nodes in the Kubernetes cluster and uses them to
    configure the MetalLB addresses. The generated values are saved to a YAML file.
    Args:
        config_folder (str): The path to the configuration folder.
        project_id (str): The project identifier.
        cluster_config_dict (dict): A dictionary containing cluster configuration details,
                                    including 'ingress_class' and 'ingress_resolver_email'.
    Returns:
        dict: A dictionary containing the namespace, repository URL, chart version, path to the
              values YAML file, release name, and chart name.
    """


    
    config.load_kube_config() 

    # Create a CoreV1Api instance
    v1 = client.CoreV1Api()

    # Get the list of nodes
    nodes = v1.list_node()

    # Extract InternalIP for each node
    internal_ips = []
    for node in nodes.items:
        for address in node.status.addresses:
            if address.type == "InternalIP":
                internal_ips.append(address.address)

    if len(internal_ips) == 1:
        internal_ips.append(internal_ips[0])

    core_toolkit_values = {
        "namespace": "maia-core-toolkit",
        "chart_version": "0.1.6",
        "repo_url": "https://kthcloud.github.io/MAIA/",
        "chart_name": "maia-core-toolkit"
    } # TODO: Change this to updated values

    core_toolkit_values.update({
        "metric_server": {
            "enabled": True
        },

        "dashboard": {
            "enabled": True
        },

        "default_ingress_class": cluster_config_dict["ingress_class"],
        "cert_manager": {
            "enabled": True,
            "email": cluster_config_dict["ingress_resolver_email"]
        },
        "metallb":{
            "enabled": True,
            "addresses": "-".join(internal_ips)
    }
    })
    

    Path(config_folder).joinpath(project_id,"core_toolkit_values").mkdir(parents=True, exist_ok=True)
    with open( Path(config_folder).joinpath(project_id,"core_toolkit_values","core_toolkit_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(core_toolkit_values))


    return {"namespace": core_toolkit_values["namespace"],
     "repo": core_toolkit_values["repo_url"],
     "version": core_toolkit_values["chart_version"],
     "values": str( Path(config_folder).joinpath(project_id,"core_toolkit_values","core_toolkit_values.yaml")),
     "release": f"{project_id}-toolkit",
     "chart": core_toolkit_values["chart_name"]
     }

def create_traefik_values(config_folder, project_id, cluster_config_dict):
    """
    Creates the Traefik values configuration file for a given project and cluster configuration.

    Args:
        config_folder (str): The path to the configuration folder.
        project_id (str): The unique identifier for the project.
        cluster_config_dict (dict): A dictionary containing the cluster configuration.

    Returns:
        dict: A dictionary containing the namespace, repository URL, chart version, values file path, release name, and chart name for the Traefik deployment.

    Raises:
        OSError: If there is an error creating the directory or writing the file.

    Example:
        config_folder = "/path/to/config"
        project_id = "my_project"
        cluster_config_dict = {
            "traefik_resolver": "myresolver",
            "ingress_resolver_email": "email@example.com",
            "domain": "example.com",
            "traefik_dashboard_password": "password"
        create_traefik_values(config_folder, project_id, cluster_config_dict)
    """
    Path(config_folder).joinpath(project_id,"traefik_values").mkdir(parents=True, exist_ok=True)

    self_signed_tls = False
    traefik_values = {
        "namespace": "traefik",
        "repo_url": "https://traefik.github.io/charts",
        "chart_name": "traefik",
        "chart_version": "33.2.1"
    } # TODO: Change this to updated values

    resolver = cluster_config_dict["traefik_resolver"]
    acme_email = cluster_config_dict["ingress_resolver_email"]
    traefik_values.update({

        "certificatesResolvers":
        {
            "maiaresolver":
            {
                "acme":
                {
                    "email": cluster_config_dict["ingress_resolver_email"],
                    "httpchallenge": "true",
                    "httpchallenge": {
                        "entryPoint": "web"
                    },
                    "caserver": "https://acme-v02.api.letsencrypt.org/directory",
                    #"caServer": "https://acme-staging-v02.api.letsencrypt.org/directory",
                    "storage": "/data/acme.json"
                }
            }
        },

"ingressRoute": {
    "dashboard": {
        "enabled": True,
        "matchRule": "Host(`{}`)".format("traefik."+cluster_config_dict["domain"]), #&& PathPrefix(`/dashboard`)
        "entryPoints": ["websecure"],
        "tls": {
            "certResolver": cluster_config_dict["traefik_resolver"]
        },
        "middlewares": [
            {"name": "traefik-dashboard-auth"},
            {"name": "traefik-dashboard-replace-path"}
        ]
    }
    },

    "extraObjects": [
    {
        "apiVersion": "v1",
        "kind": "Secret",
        "metadata": {
            "name": "traefik-dashboard-auth-secret"
        },
        "type": "kubernetes.io/basic-auth",
        "stringData": {
            "username": "admin",
            "password": cluster_config_dict["traefik_dashboard_password"]
        }
    },
    {
        "apiVersion": "traefik.io/v1alpha1",
        "kind": "Middleware",
        "metadata": {
            "name": "traefik-dashboard-replace-path"
        },
        "spec": {
            "replacePathRegex": {
                "regex": "^/dashboard/(.*)",
                "replacement": "/dashboard/$1"
            }

    }
    },
    {
        "apiVersion": "traefik.io/v1alpha1",
        "kind": "Middleware",
        "metadata": {
            "name": "traefik-dashboard-auth"
        },
        "spec": {
            "basicAuth": {
                "secret": "traefik-dashboard-auth-secret"
            }
        }
    }
],

"globalArguments": [
    "--entrypoints.web.http.redirections.entryPoint.to=:443",
    #"--entrypoints.web.http.redirections.entrypoint.to=websecure",
    "--entrypoints.web.http.redirections.entrypoint.scheme=https",
    "--global.checknewversion",
    "--global.sendanonymoususage",
    #"--entrypoints.metrics.address=:9100/tcp",
    #"--entrypoints.traefik.address=:9000/tcp",
    "--api.dashboard=true",
    "--ping=true",
    "--metrics.prometheus=true",
    "--metrics.prometheus.entrypoint=metrics",
    "--providers.kubernetescrd",
    "--providers.kubernetesingress",
    #"--entrypoints.websecure.http.tls=true",
    "--api.insecure",
    "--accesslog",
    #"--entrypoints.web.Address=:80",
    "--entryPoints.websecure.forwardedHeaders.insecure",
    "--entryPoints.web.forwardedHeaders.insecure",
    #"--entrypoints.websecure.Address=:443",
    #f"--certificatesresolvers.{resolver}.acme.httpchallenge=true",
    #f"--certificatesresolvers.{resolver}).acme.httpchallenge.entrypoint=web",
    #f"--certificatesresolvers.{resolver}.acme.email={acme_email}",
    #f"--certificatesresolvers.{resolver}.acme.storage=/data/acme.json",
    #f"--certificatesresolvers.{resolver}.acme.caserver=https://acme-v02.api.letsencrypt.org/directory"
    ]

    })


    if self_signed_tls:
        traefik_values.update({
            "tlsStore": {
                "default": {
                    "defaultCertificate": {
                        "secretName": "wildcard-domain-tls"
                    }
                }
            }
        })


    with open( Path(config_folder).joinpath(project_id,"traefik_values","traefik_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(traefik_values))

    return {"namespace": traefik_values["namespace"],
        "repo": traefik_values["repo_url"],
        "version": traefik_values["chart_version"],
        "values": str( Path(config_folder).joinpath(project_id,"traefik_values","traefik_values.yaml")),
        "release": f"{project_id}-traefik",
        "chart": traefik_values["chart_name"]
        }

def create_metallb_values(config_folder, project_id):
    """
    Creates and writes MetalLB Helm chart values to a YAML file and returns a dictionary with deployment details.

    Args:
        config_folder (str): The path to the configuration folder where the YAML file will be created.
        project_id (str): The project identifier used to create a unique directory and release name.

    Returns:
        dict: A dictionary containing the following keys:
            - "namespace" (str): The Kubernetes namespace for MetalLB.
            - "repo" (str): The URL of the MetalLB Helm chart repository.
            - "version" (str): The version of the MetalLB Helm chart.
            - "values" (str): The file path to the generated YAML values file.
            - "release" (str): The release name for the MetalLB deployment.
            - "chart" (str): The name of the MetalLB Helm chart.
    """

    metallb_values = {
        "namespace": "metallb-system",
        "chart_version": "0.14.9",
        "repo_url": "https://metallb.github.io/metallb",
        "chart_name": "metallb"
    } # TODO: Change this to updated values

    Path(config_folder).joinpath(project_id,"metallb_values").mkdir(parents=True, exist_ok=True)
    with open( Path(config_folder).joinpath(project_id,"metallb_values","metallb_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(metallb_values))

    return {"namespace": metallb_values["namespace"],
        "repo": metallb_values["repo_url"],
        "version": metallb_values["chart_version"],
        "values": str( Path(config_folder).joinpath(project_id,"metallb_values","metallb_values.yaml")),
        "release": f"{project_id}-metallb",
        "chart": metallb_values["chart_name"]
        }

def create_cert_manager_values(config_folder, project_id):
    """
    Creates a dictionary of values for configuring cert-manager and writes it to a YAML file.

    Args:
        config_folder (str): The path to the configuration folder.
        project_id (str): The project identifier.

    Returns:
        dict: A dictionary containing the namespace, repository URL, chart version, path to the values file, release name, and chart name.
    """

    cert_manager_values = { ## TODO: Causing issues from ArgoCD
        "namespace": "cert-manager",
        "chart_version": "1.16.2",
        "repo_url": "https://charts.jetstack.io",
        "chart_name": "cert-manager"
    } # TODO: Change this to updated values

    cert_manager_values.update({
        "crds": {
            "enabled": True
        },
    })

    Path(config_folder).joinpath(project_id,"cert_manager_values").mkdir(parents=True, exist_ok=True)

    with open( Path(config_folder).joinpath(project_id,"cert_manager_values","cert_manager_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(cert_manager_values))

    return {"namespace": cert_manager_values["namespace"],
        "repo": cert_manager_values["repo_url"],
        "version": cert_manager_values["chart_version"],
        "values": str( Path(config_folder).joinpath(project_id,"cert_manager_values","cert_manager_values.yaml")),
        "release": f"{project_id}-cert-manager",
        "chart": cert_manager_values["chart_name"]
        }
def create_rancher_values(config_folder, project_id, cluster_config_dict):
    """
    Generates Rancher values configuration and writes it to a YAML file.
    Args:
        config_folder (str): The path to the configuration folder.
        project_id (str): The project identifier.
        cluster_config_dict (dict): A dictionary containing cluster configuration details.
    Returns:
        dict: A dictionary containing Rancher deployment details including namespace, repo URL, chart version, values file path, release name, and chart name.
    The function performs the following steps:
    1. Initializes a dictionary with default Rancher values.
    2. Updates the dictionary with values from the cluster configuration dictionary.
    3. Adds specific annotations based on the ingress class.
    4. Creates the necessary directory structure for the configuration files.
    5. Writes the Rancher values to a YAML file.
    6. Returns a dictionary with details required for Rancher deployment.
    """
    
    rancher_values = {
        "namespace": "cattle-system",
        "repo_url": "https://releases.rancher.com/server-charts/latest",
        "chart_name": "rancher",
        "chart_version": "2.10.1"
    } # TODO: Change this to updated values

    rancher_values.update({
        "hostname": "mgmt."+cluster_config_dict["domain"],
        "ingress": {
            "extraAnnotations": {},
            "tls":
                {   
                    "source": "letsEncrypt",
                
                }
            
        },
        "letsEncrypt": {
            "email": cluster_config_dict["ingress_resolver_email"],
            "ingress": {
                "class": cluster_config_dict["ingress_class"],
                
            }
        },
        "bootstrapPassword": cluster_config_dict["rancher_password"],
    })

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        rancher_values["ingress"]["extraAnnotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        rancher_values["ingress"]["extraAnnotations"]["traefik.ingress.kubernetes.io/router.tls"] = 'true'
        rancher_values["ingress"]["extraAnnotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = cluster_config_dict["traefik_resolver"]
    elif cluster_config_dict["ingress_class"] == "nginx":
        rancher_values["ingress"]["extraAnnotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"

    Path(config_folder).joinpath(project_id,"rancher_values").mkdir(parents=True, exist_ok=True)

    with open( Path(config_folder).joinpath(project_id,"rancher_values","rancher_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(rancher_values))

    return {"namespace": rancher_values["namespace"],
        "repo": rancher_values["repo_url"],
        "version": rancher_values["chart_version"],
        "values": str( Path(config_folder).joinpath(project_id,"rancher_values","rancher_values.yaml")),
        "release": f"{project_id}-rancher",
        "chart": rancher_values["chart_name"]
        }

def create_gpu_operator_values(config_folder, project_id, cluster_config_dict):
    """
    Creates GPU operator values configuration for a Kubernetes cluster and writes it to a YAML file.
    Args:
        config_folder (str): The folder path where the configuration will be saved.
        project_id (str): The project identifier used to create a unique directory for the configuration.
        cluster_config_dict (dict): A dictionary containing cluster configuration details, including the Kubernetes distribution.
    Returns:
        dict: A dictionary containing the namespace, repository URL, chart version, path to the values file, release name, and chart name.
    """

    gpu_operator_values = {
        "namespace": "gpu-operator",
        "chart_version": "24.9.1",
        "repo_url": "https://helm.ngc.nvidia.com/nvidia",
        "chart_name": "gpu-operator"
    } # TODO: Change this to updated values
    
    if cluster_config_dict["k8s_distribution"] == "microk8s":
        gpu_operator_values["toolkit"] = {
            "env": [
                {
                    "name": "CONTAINERD_CONFIG",
                    "value": "/var/snap/microk8s/current/args/containerd-template.toml"
                },
                {
                    "name": "CONTAINERD_SOCKET",
                    "value": "/var/snap/microk8s/common/run/containerd.sock"
                },
                {
                    "name": "CONTAINERD_RUNTIME_CLASS",
                    "value": "nvidia"
                },
                {
                    "name": "CONTAINERD_SET_AS_DEFAULT",
                    "value": "true"
                }
            ]
        }

    Path(config_folder).joinpath(project_id,"gpu_operator_values").mkdir(parents=True, exist_ok=True)

    with open( Path(config_folder).joinpath(project_id,"gpu_operator_values","gpu_operator_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(gpu_operator_values))

    return {"namespace": gpu_operator_values["namespace"],
        "repo": gpu_operator_values["repo_url"],
        "version": gpu_operator_values["chart_version"],
        "values": str( Path(config_folder).joinpath(project_id,"gpu_operator_values","gpu_operator_values.yaml")),
        "release": f"{project_id}-gpu-operator",
        "chart": gpu_operator_values["chart_name"]
        }

def create_ingress_nginx_values(config_folder, project_id):
    """
    Creates and writes the ingress-nginx Helm chart values to a YAML file.

    Args:
        config_folder (str): The path to the configuration folder.
        project_id (str): The unique identifier for the project.

    Returns:
        dict: A dictionary containing the following keys:
            - "namespace" (str): The namespace for the ingress-nginx.
            - "repo" (str): The repository URL for the ingress-nginx chart.
            - "version" (str): The version of the ingress-nginx chart.
            - "values" (str): The file path to the generated ingress-nginx values YAML file.
            - "release" (str): The release name for the ingress-nginx chart.
            - "chart" (str): The name of the ingress-nginx chart.
    """

    ingress_nginx_values = {
        "namespace": "ingress-nginx",
        "repo_url": "https://kubernetes.github.io/ingress-nginx",
        "chart_name": "ingress-nginx",
        "chart_version": "4.11.3"

    } # TODO: Change this to updated values

    Path(config_folder).joinpath(project_id,"ingress_nginx_values").mkdir(parents=True, exist_ok=True)

    with open( Path(config_folder).joinpath(project_id,"ingress_nginx_values","ingress_nginx_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(ingress_nginx_values))

    return {"namespace": ingress_nginx_values["namespace"],
        "repo": ingress_nginx_values["repo_url"],
        "version": ingress_nginx_values["chart_version"],
        "values": str( Path(config_folder).joinpath(project_id,"ingress_nginx_values","ingress_nginx_values.yaml")),
        "release": f"{project_id}-ingress-nginx",
        "chart": ingress_nginx_values["chart_name"]
        }


import subprocess
from pathlib import Path
from hydra import initialize, initialize_config_dir
from hydra import compose as hydra_compose
import click
import hydra
import yaml
from omegaconf import OmegaConf
from MAIA.maia_fn import deploy_oauth2_proxy, deploy_mysql, deploy_mlflow
from MAIA.maia_admin import create_maia_namespace_values, install_maia_project, get_maia_toolkit_apps, generate_minio_configs, generate_mlfow_configs, generate_mysql_configs
from MAIA_scripts.MAIA_create_JupyterHub_config import create_jupyterhub_config_api
import datetime
import argparse
import json
import os
import asyncio
from pyhelm3 import Client
import subprocess
from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from textwrap import dedent

import MAIA
version = MAIA.__version__


TIMESTAMP = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.datetime.now())

DESC = dedent(
    """
    Script to deploy a MAIA Project Toolkit to a Kubernetes cluster. The target cluster is specified by setting the corresponding ``--cluster-config``,
    while the project-related configuration is specified with ``--project-config-file`` and ``--maia-config-file``.
    The necessary MAIA Configuration files should be found in ``--config-folder``.
    """  # noqa: E501
)
EPILOG = dedent(
    """
    Example call:
    ::
        {filename} --project-config-file /PATH/TO/form.yaml --cluster-config /PATH/TO/cluster.yaml  --config-folder /PATH/TO/config_folder --maia-config-file /PATH/TO/maia_config.yaml
    """.format(  # noqa: E501
        filename=Path(__file__).stem
    )
)

def get_arg_parser():
    pars = ArgumentParser(description=DESC, epilog=EPILOG, formatter_class=RawTextHelpFormatter)

    pars.add_argument(
        "--project-config-file",
        type=str,
        required=True,
        help="YAML configuration file used to extract the project configuration.",
    )

    pars.add_argument(
        "--maia-config-file",
        type=str,
        required=True,
        help="YAML configuration file used to extract MAIA configuration.",
    )

    pars.add_argument(
        "--cluster-config",
        type=str,
        required=True,
        help="YAML configuration file used to extract the cluster configuration.",
    )

    pars.add_argument(
        "--config-folder",
        type=str,
        required=True,
        help="Configuration Folder where to locate (and temporarily store) the MAIA configuration files.",
    )


    pars.add_argument(
        "--minimal",
        type=str,
        required=False,
        help="Optional flag to only deploy JupyterHub in the MAIA namespace.",
    )

    pars.add_argument('-v', '--version', action='version', version='%(prog)s ' + version)

    return pars


async def verify_installed_maia_toolkit(project_id, namespace, get_chart_metadata = True):

    
    client = Client(kubeconfig = os.environ["KUBECONFIG"])


    try:
        revision = await client.get_current_revision(project_id, namespace = namespace)
    except:
        print("Project not found")
        return -1
    if get_chart_metadata:
        chart_metadata = await revision.chart_metadata()
        print(
            revision.release.name,
            revision.release.namespace,
            revision.revision,
            str(revision.status),
            chart_metadata.name,
            chart_metadata.version
        )
        return revision.revision
    return {
        "revision": revision.revision,
        "release_name": revision.release.name,
        "release_namespace": revision.release.namespace,
        "status": str(revision.status)
    }

@click.command()
@click.option("--project-config-file", type=str)
@click.option("--maia-config-file", type=str)
@click.option("--cluster-config", type=str)
@click.option('--minimal', is_flag=True)
@click.option("--config-folder", type=str)
def main(project_config_file,maia_config_file, cluster_config, config_folder, minimal=False):
    deploy_maia_toolkit(project_config_file, maia_config_file, cluster_config, config_folder, minimal)

def deploy_maia_toolkit(project_config_file, maia_config_file, cluster_config, config_folder, minimal=False):
    project_form_dict = yaml.safe_load(Path(project_config_file).read_text())
    
    cluster_config_dict = yaml.safe_load(Path(cluster_config).read_text())
    maia_config_dict = yaml.safe_load(Path(maia_config_file).read_text())

    deploy_maia_toolkit_api(project_form_dict, maia_config_dict, cluster_config_dict, config_folder, minimal)

def deploy_maia_toolkit_api(project_form_dict, maia_config_dict, cluster_config_dict, config_folder,minimal=False, redeploy_enabled = False):
    group_id = project_form_dict["group_ID"]
    Path(config_folder).joinpath(project_form_dict["group_ID"]).mkdir(parents=True, exist_ok=True)

    

    namespace = project_form_dict["group_ID"].lower().replace("_", "-")



    helm_commands = []


    minio_configs = generate_minio_configs()
    mlflow_configs = generate_mlfow_configs(namespace=group_id.lower().replace("_", "-"))
    mysql_configs = generate_mysql_configs(namespace=group_id.lower().replace("_", "-"))

    project_form_dict["minio_access_key"] = minio_configs["console_access_key"]
    project_form_dict["minio_secret_key"] = minio_configs["console_secret_key"]

    helm_commands.append(create_maia_namespace_values(project_form_dict, cluster_config_dict, config_folder, minio_configs=minio_configs, mlflow_configs=mlflow_configs))


    with open(Path(config_folder).joinpath(group_id,"maia_namespace_values","namespace_values.yaml")) as f:
        maia_namespace_values = yaml.safe_load(f)
        project_form_dict["ssh_users"] = []
        for user in maia_namespace_values["users"]:
            project_form_dict["ssh_users"].append(
                {
                    "username": user["jupyterhub_username"].replace("-2d", "-").replace("-40", "@").replace("-2e", "."),
                    "ssh_port": user["sshPort"]
                }
            )

    helm_commands.append(create_jupyterhub_config_api(
        project_form_dict,
        maia_config_dict,
        cluster_config_dict,
        config_folder
    ))


    helm_commands.append(deploy_oauth2_proxy(cluster_config_dict, project_form_dict, config_folder))

    helm_commands.append(deploy_mysql(cluster_config_dict, project_form_dict, config_folder, mysql_configs=mysql_configs))
    helm_commands.append(deploy_mlflow(cluster_config_dict, project_form_dict, config_folder,mysql_config=mysql_configs, minio_config=minio_configs))



    for helm_command in helm_commands:
        cmd = [
            "helm",
            "upgrade",
            "--install",
            "--wait",
            "-n",
            helm_command["namespace"],
            helm_command["release"],
            helm_command["chart"],
            "--repo",
            helm_command["repo"],
            "--version",
            helm_command["version"],
            "--values",
            helm_command["values"],
        ]
        print(" ".join(cmd))

    destination_cluster_address = cluster_config_dict["argocd_destination_cluster_address"]

    values = {
       "defaults": [
              "_self_",
              {"maia_namespace_values": "namespace_values"},
              {"jupyterhub_values": "jupyterhub_values"},
            {"oauth2_proxy_values": "oauth2_proxy_values"},
            {"mysql_values": "mysql_values"},
            {"mlflow_values": "mlflow_values"}
         ],
        "argo_namespace": maia_config_dict["argocd_namespace"],
        "group_ID": f"MAIA:{group_id}",
        "destination_server": f"{destination_cluster_address}",
        "sourceRepos": [
            "https://kthcloud.github.io/MAIA/",
            "https://hub.jupyter.org/helm-chart/",
            "https://oauth2-proxy.github.io/manifests"
        ]
    }
    
    with open(Path(config_folder).joinpath(group_id,"values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(values))

    try:
        initialize_config_dir(config_dir=str(Path(config_folder).joinpath(group_id)), job_name=group_id)
    except:
        hydra.core.global_hydra.GlobalHydra.instance().clear()
        initialize_config_dir(config_dir=str(Path(config_folder).joinpath(group_id)), job_name=group_id)
    cfg = hydra_compose("values.yaml")
    OmegaConf.save(cfg, str(Path(config_folder).joinpath(group_id, f"{group_id}_values.yaml")), resolve=True)

    project_id = namespace
    
    

    revision = asyncio.run(verify_installed_maia_toolkit(project_id, maia_config_dict["argocd_namespace"]))

    if revision == -1 or redeploy_enabled:
        print("Installing MAIA Workspace")
    
        
        project_chart = maia_config_dict["maia_project_chart"]
        project_repo = maia_config_dict["maia_project_repo"]
        project_version = maia_config_dict["maia_project_version"]
        asyncio.run(install_maia_project(group_id, Path(config_folder).joinpath(group_id, f"{group_id}_values.yaml"),maia_config_dict["argocd_namespace"], project_chart, project_repo=project_repo, project_version=project_version))
    else:
        argocd_host = maia_config_dict["argocd_host"]
        token = maia_config_dict["argocd_token"]
        print("MAIA Workspace already installed")
        asyncio.run(get_maia_toolkit_apps(group_id, token, argocd_host))
if __name__ == "__main__":
    main()



from omegaconf import OmegaConf
from pathlib import Path
from MAIA.maia_fn import encode_docker_registry_secret, convert_username_to_jupyterhub_username, get_ssh_ports
import argocd_client
from argocd_client.rest import ApiException
from pyhelm3 import Client
from pprint import pprint
import os
import yaml
from secrets import token_urlsafe
import base64


def generate_minio_configs():
    """
    Generate configuration settings for MinIO.

    This function generates a dictionary containing access keys and secret keys
    for MinIO, including base64 encoded keys for console access.

    Returns:
        dict: A dictionary with the following keys:
            - access_key (str): The access key for MinIO.
            - secret_key (str): A randomly generated secret key for MinIO.
            - console_access_key (str): A base64 encoded access key for console access.
            - console_secret_key (str): A base64 encoded secret key for console access.
    """

    minio_configs = {
        "access_key": "admin",
        "secret_key": token_urlsafe(16).replace("-","_"),
        "console_access_key": base64.b64encode(token_urlsafe(16).replace("-","_").encode("ascii")).decode("ascii"),
        "console_secret_key": base64.b64encode(token_urlsafe(16).replace("-","_").encode("ascii")).decode("ascii")
    }

    return minio_configs

def generate_mlfow_configs(namespace):
    """
    Generate MLflow configuration dictionary with encoded user and password.
    Args:
        namespace (str): The namespace to be encoded as the MLflow user.
    Returns:
        dict: A dictionary containing the encoded MLflow user and password.
    """
    
    mlfow_configs = {
        "mlflow_user": base64.b64encode(namespace.encode("ascii")).decode("ascii"),
        "mlflow_password": base64.b64encode(token_urlsafe(16).replace("-","_").encode("ascii")).decode("ascii"),
    }

    return mlfow_configs

def generate_mysql_configs(namespace):
    """
    Generate MySQL configuration dictionary.

    Args:
        namespace (str): The namespace to be used as the MySQL user.

    Returns:
        dict: A dictionary containing MySQL user and password.
    """
    mysql_configs = {
        "mysql_user": namespace,
        "mysql_password": token_urlsafe(16),
    }

    return mysql_configs

def create_maia_namespace_values(namespace_config, cluster_config, config_folder, minio_configs=None, mlflow_configs=None):
    """
    Create MAIA namespace values for deployment.
    Args:
        namespace_config (dict): Configuration for the namespace, including group ID and users.
        cluster_config (dict): Configuration for the cluster, including SSH port type, port range, and storage class.
        config_folder (str): Path to the folder where configuration files will be saved.
        minio_configs (dict, optional): Configuration for MinIO, including access keys and console keys. Defaults to None.
        mlflow_configs (dict, optional): Configuration for MLflow, including user and password. Defaults to None.
    Returns:
        dict: A dictionary containing namespace, release, chart, repo, version, and values file path.
    """

    group_id = namespace_config["group_ID"]
    users = []

    ssh_ports = get_ssh_ports(
        n_requested_ports=len(namespace_config["users"]),
        port_type=cluster_config["ssh_port_type"],
        ip_range=cluster_config["port_range"],
        
    )

    if ssh_ports is None:
        for user in namespace_config["users"]:
            users.append({
                "jupyterhub_username": convert_username_to_jupyterhub_username(user),
                "sshPort": None
            })
    
    else:
        for ssh_port, user in zip(ssh_ports, namespace_config["users"]):
            users.append({
                "jupyterhub_username": convert_username_to_jupyterhub_username(user),
                "sshPort": ssh_port
                })
        

    maia_namespace_values = {
        "pvc": {
            "pvc_type": cluster_config["shared_storage_class"],
            "access_mode": "ReadWriteMany",
            "size": "10Gi"
        },
        "chart_name": "maia-namespace", 
        "chart_version": "0.1.6", 
        "repo_url": "https://kthcloud.github.io/MAIA/", 
        "namespace": namespace_config["group_ID"].lower().replace("_", "-"),
        "serviceType": cluster_config["ssh_port_type"],
        "users": users,
        "metallbSharedIp": cluster_config.get("metallb_shared_ip",False),
        "metallbIpPool": cluster_config.get("metallb_ip_pool",False),
        "loadBalancerIp": cluster_config.get("maia_metallb_ip",False),
        
    } # TODO: Change this to updated values
    if "imagePullSecrets" in cluster_config:
        maia_namespace_values["dockerRegistrySecret"] = {
            "enabled": True,
            "dockerRegistrySecretName": cluster_config["imagePullSecrets"],
            "dockerRegistrySecret": encode_docker_registry_secret(
                cluster_config["docker_server"],
                cluster_config["docker_username"],
                cluster_config["docker_password"]
            )
        }



    if minio_configs:
        maia_namespace_values["minio"] = {
            "enabled": True,
            "consoleDomain": "https://{}.{}/minio-console".format(namespace_config["group_subdomain"],cluster_config["domain"]),
            "namespace": namespace_config["group_ID"].lower().replace("_", "-"),
            "storageClassName": cluster_config["storage_class"],
            "storageSize": "10Gi",
            "accessKey": minio_configs["access_key"],
            "secretKey": minio_configs["secret_key"],
            "clientId": cluster_config["keycloak"]["client_id"],
            "clientSecret": cluster_config["keycloak"]["client_secret"],
            "openIdConfigUrl": cluster_config["keycloak"]["issuer_url"] + "/.well-known/openid-configuration",
            "consoleAccessKey": minio_configs["console_access_key"],
            "consoleSecretKey": minio_configs["console_secret_key"]
        }

        if cluster_config["url_type"] == "subpath":
            maia_namespace_values["minio"]["consoleDomain"] = "https://{}/{}-minio-console".format(
                cluster_config["domain"], namespace_config["group_ID"].lower().replace("_", "-"))
 
    if mlflow_configs:
        maia_namespace_values["mlflow"] = {
            "enabled": True,
            "user": mlflow_configs["mlflow_user"],
            "password": mlflow_configs["mlflow_password"],
        }

    namespace_id = namespace_config["group_ID"].lower().replace("_", "-")
    Path(config_folder).joinpath(group_id,"maia_namespace_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(group_id,"maia_namespace_values","namespace_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(maia_namespace_values))


    return {
        "namespace": maia_namespace_values["namespace"],
        "release": f"{namespace_id}-namespace",
        "chart": maia_namespace_values["chart_name"],
        "repo": maia_namespace_values["repo_url"],
        "version": maia_namespace_values["chart_version"],  
        "values": str(Path(config_folder).joinpath(group_id,"maia_namespace_values","namespace_values.yaml"))
    }


async def get_maia_toolkit_apps(group_id, token, argo_cd_host):
    """
    Retrieve and print information about a specific project and its associated applications from Argo CD.
    Args:
        group_id (str): The group identifier used to construct project and application names.
        token (str): The authorization token for accessing the Argo CD API.
        argo_cd_host (str): The host URL of the Argo CD server.
    Returns:
        None
    Raises:
        ApiException: If there is an error when calling the Argo CD API.
    """
    configuration = argocd_client.Configuration(
        host = argo_cd_host

    )
    with argocd_client.ApiClient(configuration) as api_client:
        # Add the Authorization header
        api_client.default_headers["Authorization"] = f"Bearer {token}"

        # Create instances of the RepositoryServiceApi and ApplicationServiceApi


        api_instance = argocd_client.ProjectServiceApi(api_client)
        api_instance_apps = argocd_client.ApplicationServiceApi(api_client)
        name = group_id.lower().replace("_","-") # str | 

        try:
            # Get returns a project by name
            api_response = api_instance.get_mixin6(name)
            pprint(api_response)
        except ApiException as e:
            print("Exception when calling ProjectServiceApi->get_mixin6: %s\n" % e)


        names = [
                    f'{group_id.lower().replace("_","-")}-namespace',
                    f'{group_id.lower().replace("_","-")}-jupyterhub',
                    f'{group_id.lower().replace("_","-")}-oauth2-proxy',
                ]  # TODO: Add more apps
        project = [group_id.lower().replace("_","-")]
        for name in names:
            try:
            # Get returns an application by name
                
                #body = argocd_client.ApplicationApplicationSyncRequest()
                #api_response = api_instance_apps.sync(name, body)
                #api_response = api_instance_apps.delete_mixin9(name, cascade=True)
                #pprint(api_response)
                api_response = api_instance_apps.get_mixin9(name, project=project)
                pprint(api_response)
            except ApiException as e:
                print("Exception when calling ApplicationServiceApi->get_mixin9: %s\n" % e)

async def install_maia_project(group_id, values_file, argo_cd_namespace, project_chart, project_repo=None, project_version=None):
    """
    Installs or upgrades a MAIA project using the specified Helm chart and values file.
    Args:
        group_id (str): The group ID for the project. This will be used as the release name.
        values_file (str): Path to the YAML file containing the values for the Helm chart.
        argo_cd_namespace (str): The namespace in which to install the project.
        project_chart (str): The name of the Helm chart to use for the project.
        project_repo (str, optional): The repository URL where the Helm chart is located. Defaults to None.
        project_version (str, optional): The version of the Helm chart to use. Defaults to None.
    Returns:
        None
    Raises:
        FileNotFoundError: If the values file does not exist.
        yaml.YAMLError: If there is an error parsing the values file.
        Exception: If there is an error during the installation or upgrade process.
    Example:
        await install_maia_project(
            group_id="example_group",
            values_file="/path/to/values.yaml",
            argo_cd_namespace="default",
            project_chart="example_chart",
            project_repo="https://example.com/charts",
            project_version="1.0.0"
    """

          
    client = Client(kubeconfig = os.environ["KUBECONFIG"])


    chart = await client.get_chart(
        project_chart,
        repo = project_repo,
        version = project_version
    )

    with open(values_file) as f:
        values = yaml.safe_load(f)
    
    
    revision = await client.install_or_upgrade_release(
        group_id.lower().replace("_","-"),
        chart,
        values,
        namespace=argo_cd_namespace,
        wait = True
    )
    print(
        revision.release.name,
        revision.release.namespace,
        revision.revision,
        str(revision.status)
    )


def create_maia_admin_toolkit_values(config_folder, project_id, cluster_config_dict, maia_config_dict):
    """
    Creates and writes the MAIA admin toolkit values to a YAML file.
    Args:
        config_folder (str): The path to the configuration folder.
        project_id (str): The project identifier.
        cluster_config_dict (dict): Dictionary containing cluster configuration values.
        maia_config_dict (dict): Dictionary containing MAIA configuration values.
    Returns:
        dict: A dictionary containing the namespace, release name, chart name, repository URL, chart version, 
              and the path to the generated values YAML file.
    """
    
    admin_group_ID = maia_config_dict["admin_group_ID"]

    admin_toolkit_values = {
        "namespace": "maia-admin-toolkit",
        "repo_url": "https://kthcloud.github.io/MAIA/",
        "chart_name": "maia-admin-toolkit",
        "chart_version": "1.1.0",
    } # TODO: Change this to updated values

    admin_toolkit_values.update(
        {   
            "argocd":{
                "enabled": True,
                "argocd_namespace": "argocd",
                "argocd_domain": "argocd."+cluster_config_dict["domain"],
                "keycloak_issuer_url": "https://iam."+cluster_config_dict["domain"]+"/realms/maia",
                "keycloak_client_id": "maia",
                "keycloak_client_secret": cluster_config_dict["keycloak_maia_client_secret"],
            },
            "certResolver": cluster_config_dict["traefik_resolver"],
            "admin_group_ID": admin_group_ID,
            "harbor": {
                "enabled": True,
                "values": {
                    "namespace": "harbor",
                    "storageClassName": cluster_config_dict["storage_class"],
                }
            },
            "dashboard": {
                "enabled": True,
                "dashboard_domain": "dashboard."+cluster_config_dict["domain"],
            }

        }
    )

    Path(config_folder).joinpath(project_id,"maia_admin_toolkit_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id,"maia_admin_toolkit_values","maia_admin_toolkit_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(admin_toolkit_values))

    return {
        "namespace": admin_toolkit_values["namespace"],
        "release": f"{project_id}-toolkit",
        "chart": admin_toolkit_values["chart_name"],
        "repo": admin_toolkit_values["repo_url"],
        "version": admin_toolkit_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id,"maia_admin_toolkit_values","admin_toolkit_values.yaml"))
    }



def create_harbor_values(config_folder, project_id, cluster_config_dict):
    """
    Create and save Harbor values configuration for a given project and cluster configuration.

    Args:
        config_folder (str): The path to the configuration folder where the Harbor values file will be saved.
        project_id (str): The unique identifier for the project.
        cluster_config_dict (dict): A dictionary containing cluster configuration details, including:
            - domain (str): The domain name for the Harbor registry.
            - ingress_class (str): The ingress class to be used (e.g., "maia-core-traefik", "nginx").
            - traefik_resolver (str, optional): The Traefik resolver to be used if ingress_class is "maia-core-traefik".

    Returns:
        dict: A dictionary containing the following keys:
            - namespace (str): The Kubernetes namespace for Harbor.
            - release (str): The release name for the Harbor Helm chart.
            - chart (str): The name of the Harbor Helm chart.
            - repo (str): The URL of the Harbor Helm chart repository.
            - version (str): The version of the Harbor Helm chart.
            - values (str): The path to the generated Harbor values YAML file.
    """

    DOMAIN = cluster_config_dict["domain"]
    harbor_values = {
       "namespace": "harbor",
       "repo_url": "https://helm.goharbor.io",
         "chart_name": "harbor",
            "chart_version": "1.16.0",
    } # TODO: Change this to updated values


    harbor_values.update({"expose": {
        "type": "ingress",
        "tls": {
            "enabled": True,
        },
        "ingress": {
            "hosts": {
                "core": f"registry.{DOMAIN}"
            },
            "annotations": {
            },
            "controller": "default",
            "className": cluster_config_dict["ingress_class"]
        }
    },
    "externalURL": f"https://registry.{DOMAIN}",
    "persistence": {
        "enabled": True,
        "resourcePolicy": "keep",
        "persistentVolumeClaim": {
            "registry": {
                "existingClaim": "pvc-harbor",
                "subPath": "registry",
                "storageClass": cluster_config_dict["ingress_class"],
                "accessMode": "ReadWriteMany"
            },
            "jobservice": {
                "jobLog": {
                    "existingClaim": "pvc-harbor",
                    "subPath": "job_logs",
                    "storageClass": cluster_config_dict["ingress_class"],
                    "accessMode": "ReadWriteMany"
                }
            },
            "database": {
                "existingClaim": "pvc-harbor",
                "subPath": "database",
                "storageClass": cluster_config_dict["ingress_class"],
                "accessMode": "ReadWriteMany"
            },
            "redis": {
                "existingClaim": "pvc-harbor",
                "subPath": "redis",
                "storageClass": cluster_config_dict["ingress_class"],
                "accessMode": "ReadWriteMany"
            },
            "trivy": {
                "existingClaim": "pvc-harbor",
                "subPath": "trivy",
                "storageClass": cluster_config_dict["ingress_class"],
                "accessMode": "ReadWriteMany"
            }
        },
        "imageChartStorage": {
            "type": "filesystem"
        }
    },
    "database": {
        "internal": {
            "password": "harbor"
        }
    },
    "metrics": {
        "enabled": True,
        "core": {
            "path": "/metrics",
            "port": 8001
        },
        "registry": {
            "path": "/metrics",
            "port": 8001
        },
        "jobservice": {
            "path": "/metrics",
            "port": 8001
        },
        "exporter": {
            "path": "/metrics",
            "port": 8001
        }
    }
    })

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        harbor_values["expose"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        harbor_values["expose"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = 'true'
        harbor_values["expose"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = cluster_config_dict["traefik_resolver"]
    elif cluster_config_dict["ingress_class"] == "nginx":
        harbor_values["expose"]["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"



    Path(config_folder).joinpath(project_id,"harbor_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id,"harbor_values","harbor_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(harbor_values))

    return {
        "namespace": harbor_values["namespace"],
        "release": f"{project_id}-harbor",
        "chart": harbor_values["chart_name"],
        "repo": harbor_values["repo_url"],
        "version": harbor_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id,"harbor_values","harbor_values.yaml"))
    }

def create_keycloak_values(config_folder, project_id, cluster_config_dict):
    """
    Generates Keycloak Helm chart values and writes them to a YAML file.
    Args:
        config_folder (str): The path to the configuration folder where the YAML file will be saved.
        project_id (str): The project identifier used to create a unique namespace and release name.
        cluster_config_dict (dict): A dictionary containing cluster configuration details such as domain, ingress class, and traefik resolver.
    Returns:
        dict: A dictionary containing the namespace, release name, chart name, repository URL, chart version, and the path to the generated values YAML file.
    """


    keycloak_values = {
        "namespace": "keycloak",
        "repo_url": "https://charts.bitnami.com/bitnami",
        "chart_name": "keycloak",
        "chart_version": "24.2.0",
    } # TODO: Change this to updated values

    keycloak_values.update(

        {    
            
            
            "extraEnvVars": [
            #{
            #    "name": "KC_HOSTNAME",
            #    "value": "keycloak."+cluster_config_dict["domain"],
            #}
            #{  "name": "KC_HOSTNAME_ADMIN_URL",
            #   "value":"keycloak."+cluster_config_dict["domain"]},
            #{  "name": "KC_HOSTNAME_URL",
            #    "value":"keycloak."+cluster_config_dict["domain"]},
            {
                "name": "KEYCLOAK_EXTRA_ARGS",
                "value": "--import-realm"
            },
        {  "name": "PROXY_ADDRESS_FORWARDING",
            "value": "true"},
        {  "name": "KEYCLOAK_HOSTNAME",
            "value": "iam."+cluster_config_dict["domain"]},

            ],
        "proxy": "edge",
        "ingress": {
            "enabled": True,
            "tls": True,
            "ingressClassName": cluster_config_dict["ingress_class"],
            "hostname": "iam."+cluster_config_dict["domain"],
            "annotations": {
                
            },
            #"secrets": [
            #    {
            #        "name": "keycloak."+cluster_config_dict["domain"]+"-tls",
            #        "key": "",
            #        "certificate": ""
            #    }
            #]
            
        },
        "extraVolumeMounts": [
            {
                "name": "keycloak-import",
                "mountPath": "/opt/bitnami/keycloak/data/import"
            }
        ],
        "extraVolumes": [
            {
                "name": "keycloak-import",
                "configMap": {
                    "name": "maia-realm-import"
                }
            }]
       # "extraVolumes": [
       #     {
       #         "name": "keycloak-export",
       #         "persistentVolumeClaim": {
       #             "claimName": "pvc-keycloak-export"
       #         }
       #     }]
    })

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        keycloak_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        keycloak_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = 'true'
        keycloak_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = cluster_config_dict["traefik_resolver"]
    elif cluster_config_dict["ingress_class"] == "nginx":
        keycloak_values["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"

    Path(config_folder).joinpath(project_id,"keycloak_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id,"keycloak_values","keycloak_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(keycloak_values))

    return {
        "namespace": keycloak_values["namespace"],
        "release": f"{project_id}-keycloak",
        "chart": keycloak_values["chart_name"],
        "repo": keycloak_values["repo_url"],
        "version": keycloak_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id,"keycloak_values","keycloak_values.yaml"))
    }

def create_loginapp_values(config_folder, project_id, cluster_config_dict):
    """
    Creates and writes the loginapp values configuration file for a given project and cluster configuration.
    Args:
        config_folder (str): The base directory where the configuration files will be stored.
        project_id (str): The unique identifier for the project.
        cluster_config_dict (dict): A dictionary containing cluster configuration details, including:
            - keycloak_maia_client_secret (str): The client secret for Keycloak.
            - domain (str): The domain name for the cluster.
            - ingress_class (str): The ingress class to be used (e.g., "maia-core-traefik" or "nginx").
            - traefik_resolver (str, optional): The Traefik resolver to be used if ingress_class is "maia-core-traefik".
    Returns:
        dict: A dictionary containing the namespace, release name, chart name, repository URL, chart version, and the path to the generated values file.
    Raises:
        KeyError: If required keys are missing from the cluster_config_dict.
        OSError: If there is an error creating directories or writing the configuration file.
    """
    loginapp_values = {
        "namespace": "authentication",
        "repo_url": "https://storage.googleapis.com/loginapp-releases/charts/",
        "chart_name": "loginapp",
        "chart_version": "1.3.0",
    } # TODO: Change this to updated values


    secret = token_urlsafe(16).replace("-","_")
    client_id = "maia"
    client_secret = cluster_config_dict["keycloak_maia_client_secret"]
    issuer_url = "https://iam."+cluster_config_dict["domain"]+"/realms/maia"
    cluster_server_address = "https://"+cluster_config_dict["domain"]+":16443"
    ca_file = "/var/snap/microk8s/current/certs/ca.crt"

    loginapp_values.update(
        {   
            "env": {
                "LOGINAPP_NAME": "MAIA Login"
            },

            "configOverwrites": {
                "oidc": {
                    "scopes": ["openid","profile","email"]
                },
                "service": {
                    "type": "ClusterIP"
                }
            },
            "ingress": {
                "enabled": True,
                "annotations": {
                },
                "tls": [
                    {
                        "hosts": [
                            "login."+cluster_config_dict["domain"]
                        ]
                    }
                ],
                "hosts": [
                    {
                        "host": "login."+cluster_config_dict["domain"],
                        "paths": [
                            {
                                "path": "/",
                                "pathType": "Prefix"
                            }
                        ]
                    }
                ]},
            "config": {
                "tls": {
                    "enabled": False
                },
                "issuerInsecureSkipVerify": True,
                "refreshToken": True,
                "clientRedirectURL": f"https://login."+cluster_config_dict["domain"]+"/callback",
                "secret": secret,
                "clientID": client_id,
                "clientSecret": client_secret,
                "issuerURL": issuer_url,
                "clusters": [
                    {
                        "server": cluster_server_address,
                        "name": "MAIA",
                        "insecure-skip-tls-verify": True,
                        "certificate-authority": ca_file
                    }
                ]
                
            }
        }
    )

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        loginapp_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        loginapp_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = 'true'
        loginapp_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = cluster_config_dict["traefik_resolver"]
    elif cluster_config_dict["ingress_class"] == "nginx":
        loginapp_values["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"
        loginapp_values["ingress"]["tls"][0]["secretName"] = "loginapp."+cluster_config_dict["domain"]

    Path(config_folder).joinpath(project_id,"loginapp_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id,"loginapp_values","loginapp_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(loginapp_values))

    return {
        "namespace": loginapp_values["namespace"],
        "release": f"{project_id}-loginapp",
        "chart": loginapp_values["chart_name"],
        "repo": loginapp_values["repo_url"],
        "version": loginapp_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id,"loginapp_values","loginapp_values.yaml"))
    }


def create_minio_operator_values(config_folder, project_id, cluster_config_dict):
    """
    Creates and writes MinIO operator values to a YAML file and returns a dictionary with deployment details.

    Args:
        config_folder (str): The path to the configuration folder.
        project_id (str): The unique identifier for the project.
        cluster_config_dict (dict): A dictionary containing cluster configuration details.

    Returns:
        dict: A dictionary containing the namespace, release name, chart name, repository URL, chart version, 
              and the path to the generated YAML values file.
    """

    minio_operator_values = {
        "namespace": "minio-operator",
        "repo_url": "https://operator.min.io",
        "chart_name": "operator",
        "chart_version": "6.0.4",
    } # TODO: Change this to updated values


    Path(config_folder).joinpath(project_id,"minio_operator_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id,"minio_operator_values","minio_operator_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(minio_operator_values))

    return {
        "namespace": minio_operator_values["namespace"],
        "release": f"{project_id}-minio-operator",
        "chart": minio_operator_values["chart_name"],
        "repo": minio_operator_values["repo_url"],
        "version": minio_operator_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id,"minio_operator_values","minio_operator_values.yaml"))
    }
